#ifndef EFFECT_H
#define EFFECT_H

#include <vd2/system/refcount.h>
#include <vd2/system/vdstl.h>
#include <map>

enum {
	kVDShaderOp_nop,
	kVDShaderOp_mov,
	kVDShaderOp_add,
	kVDShaderOp_sub,
	kVDShaderOp_mul,
	kVDShaderOp_mad,
	kVDShaderOp_dp2,
	kVDShaderOp_dp3,
	kVDShaderOp_dp4,
	kVDShaderOp_rsq,
	kVDShaderOp_rcp,
	kVDShaderOp_texld
};

enum {
	kVDShaderRegR0 = 0x1000,
	kVDShaderRegC0 = 0x2000,		// fixed constants
	kVDShaderRegK0 = 0x3000,		// varying constants
	kVDShaderRegO0 = 0x4000,		// outputs
	kVDShaderRegS0 = 0x5000,		// sampler
	kVDShaderRegGS0 = 0x8000,		// global sampler
	kVDShaderRegIndexMask = 0x0FFF,
	kVDShaderRegTypeMask = 0xF000
};

enum {
	kVDShaderMod_Negate = 0x01
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderConstant {
	float v[4];

	bool operator==(const VDShaderConstant& c) const;
};

struct VDShaderConstantHash {
	size_t operator()(const VDShaderConstant& c) const;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderInsnSource {
	int mRegister;
	unsigned char mSwizzle;
	unsigned char mModifiers;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderInsn {
	int mOp;
	int mDestRegister;
	unsigned char mWriteMask;
	VDShaderInsnSource mSources[4];
};

///////////////////////////////////////////////////////////////////////////////
enum VDShaderFilterMode {
	kFilterNone,
	kFilterPoint,
	kFilterBilinear
};

struct VDShaderSamplerState {
	bool	mbAddressUWrap : 1;
	bool	mbAddressUWrapSet : 1;
	bool	mbAddressVWrap : 1;
	bool	mbAddressVWrapSet : 1;
	bool	mbMinFilter : 1;
	bool	mbMinFilterSet : 1;
	bool	mbMagFilter : 1;
	bool	mbMagFilterSet : 1;
	bool	mbMipFilter : 1;
	bool	mbMipFilterSet : 1;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderVariable {
	int mNameToken;
	int mSemanticToken;
	int mRegister;
	int mType;
	bool mbUniform : 1;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderFunction {
	VDShaderFunction();
	~VDShaderFunction();

	typedef vdfastvector<VDShaderVariable> Parameters;
	Parameters mParameters;
	int mParameterRegisterCount;
	int mRegisterCount;
	int mSamplerCount;
	int mReturnType;
	int mReturnSemantic;
	int mNameToken;

	typedef vdfastvector<VDShaderConstant> Constants;
	Constants mConstants;

	typedef vdfastvector<VDShaderInsn> Insns;
	Insns mInsns;

	typedef vdfastvector<int> SamplerBindings;
	SamplerBindings mSamplerBindings;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderNameTable {
public:
	VDShaderNameTable();
	~VDShaderNameTable();

	int AddName(const char *s, size_t len);

	int GetName(const char *s, size_t len) const;

	const char *GetName(int id) const {
		return &mNameHeap[id];
	}

protected:
	typedef vdfastvector<char> NameHeap;
	NameHeap mNameHeap;

	typedef std::multimap<unsigned, int> NameMap;
	NameMap mNameMap;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderPass {
public:
	VDShaderPass();
	~VDShaderPass();

	int GetNameToken() const { return mNameToken; }
	int GetPixelShaderIndex() const { return mPixelShaderIndex; }

protected:
	friend class VDShaderParser;

	int mNameToken;
	int mPixelShaderIndex;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderTechnique {
public:
	VDShaderTechnique();
	~VDShaderTechnique();

	int GetNameToken() const { return mNameToken; }
	int GetPassCount() const { return mPasses.size(); }
	VDShaderPass *GetPassByIndex(int index) const {
		return (unsigned)index < mPasses.size() ? mPasses[index] : NULL;
	}

protected:
	friend class VDShaderParser;

	int mNameToken;

	typedef vdfastvector<VDShaderPass *> Passes;
	Passes mPasses;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderEffect : public vdrefcounted<IVDRefCount> {
public:
	VDShaderEffect();
	~VDShaderEffect();

	const VDShaderNameTable& NameTable() const { return mNameTable; }

	int GetFunctionCount() const { return mFunctions.size(); }
	int GetTechniqueCount() const { return mTechniques.size(); }
	int GetVariableCount() const { return mVariables.size(); }

	const VDShaderFunction *GetFunctionByIndex(int index) const {
		return (unsigned)index < mFunctions.size() ? mFunctions[index] : NULL;
	}

	const VDShaderTechnique *GetTechniqueByIndex(int index) const {
		return (unsigned)index < mTechniques.size() ? mTechniques[index] : NULL;
	}

	const VDShaderVariable *GetVariableByIndex(int index) const {
		return (unsigned)index < mVariables.size() ? &mVariables[index] : NULL;
	}

protected:
	friend class VDShaderParser;

	VDShaderNameTable mNameTable;

	typedef vdfastvector<VDShaderFunction *> Functions;
	Functions mFunctions;

	typedef vdfastvector<VDShaderTechnique *> Techniques;
	Techniques mTechniques;

	typedef vdfastvector<VDShaderVariable> Variables;
	Variables mVariables;
};

struct VDShaderBindingElement {
	int mIndex;
	int mTypeToken;
	int mSemanticToken;
};

#endif
