//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <new>
#include <vd2/system/math.h>
#include <vd2/system/registry.h>
#include <vd2/system/strutil.h>
#include <vd2/system/VDString.h>
#include <vd2/plugin/vdvideofiltold.h>
#include "vdfilter.h"
#include "effect.h"
#include "parser.h"
#include "interp.h"
#include "log.h"
#include "jit_scalar.h"
#include "editor.h"

int g_VDFilterAPIVersion;

class VDShaderFilter : public VDFilter<VDShaderFilter> {
public:
	VDShaderFilter(FilterActivation *fa, const FilterFunctions *ff);

	long GetParams();
	void Start();
	void Stop();
	bool Run();
	bool Configure(HWND hwnd);
	int Serialize(char *buf, int maxbuf);
	int Deserialize(const char *buf, int maxbuf);
	void GetSettingString(char *buf, int maxlen);
	void GetScriptString(char *buf, int maxlen);

	static CScriptObject sScriptObject[];
	static ScriptFunctionDef sFuncList[];

protected:
	static void ScriptConfig(IScriptInterpreter *, void *, const CScriptValue *, int);

	uint32		*mpTexture;
	ptrdiff_t	mTexturePitch;

	vdrefptr<VDShaderEffect> mpEffect;
	VDShaderInterpreterSampler	mSampler;

	VDShaderInterpreter	*mpInterp;
	VDShaderScalarJIT	*mpJIT;

	VDShaderFilterConfig	mConfig;

	IVDShaderFilterSource	*mpEditorSource;
};

CScriptObject VDShaderFilter::sScriptObject[] = {
	NULL, sFuncList, NULL
};

ScriptFunctionDef VDShaderFilter::sFuncList[] = {
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0si" },
	{ NULL, NULL, NULL }
};

VDShaderFilter::VDShaderFilter(FilterActivation *fa, const FilterFunctions *ff)
	: VDFilter(fa, ff)
	, mpTexture(NULL)
	, mpInterp(NULL)
	, mpJIT(NULL)
{
	mConfig.mFilename[0] = 0;
}

long VDShaderFilter::GetParams() {
	return 0;
}

namespace {
	class VDShaderParserBufOutput : public IVDShaderLogOutput {
	public:
		const char *c_str() const { return &mBuffer[0]; }

		void WriteLogOutput(const char *s) {
			mBuffer.insert(mBuffer.end(), s, s+strlen(s));
		}

		void End() {
			mBuffer.push_back(0);
		}

	protected:
		std::vector<char> mBuffer;
	};
}

void VDShaderFilter::Start() {
	uint32 srcw = fa->src.w;
	uint32 srch = fa->src.h;

	mpTexture = new(std::nothrow) uint32[(srcw+2) * (srch+2)];
	if (!mpTexture)
		ff->ExceptOutOfMemory();

	mTexturePitch = (srcw + 2) * sizeof(uint32);

	if (mConfig.mpEffect) {
		mpEffect = mConfig.mpEffect;
	} else {
		VDStringA buf;

		if (mConfig.mpSource)
			mConfig.mpSource->ReadSource(buf);
		else {
			FILE *f = fopen(mConfig.mFilename, "rb");
			if (!f)
				ff->Except("Unable to open shader file:\n%s", mConfig.mFilename);

			fseek(f, 0, SEEK_END);
			int len = ftell(f);
			buf.resize(len);
			fseek(f, 0, SEEK_SET);
			fread(&buf[0], len, 1, f);
			fclose(f);
		}

		VDShaderParser parser;
		VDShaderParserBufOutput out;
		parser.Parse(buf.data(), buf.size(), mConfig.mFilename, &out, ~mpEffect);

		if (!mpEffect) {
			out.End();
			ff->Except("Shader failed to compile:\n%s", out.c_str());
		}
	}

	if (!mpEffect->GetTechniqueCount())
		ff->Except("No techniques were declared in the .fx file.");

	const VDShaderTechnique *tech = mpEffect->GetTechniqueByIndex(0);

	if (!tech->GetPassCount())
		ff->Except("No passes were declared in the technique \"%s.\"", mpEffect->NameTable().GetName(tech->GetNameToken()));

	const VDShaderPass *pass = tech->GetPassByIndex(0);
	int psindex = pass->GetPixelShaderIndex();

	if (!psindex)
		ff->Except("No pixel shader declared in pass.");

	const VDShaderFunction *func = mpEffect->GetFunctionByIndex(psindex);

	const VDShaderBindingElement kElements[]={
		0,
		nsVDShaderParser::kTypeFloat2,
		mpEffect->NameTable().GetName("COLOR0", 6)
	};

	mSampler.Init(vdptroffset(mpTexture, mTexturePitch) + 1, mTexturePitch, srcw, srch, true, true, false);

	switch(mConfig.mExecutionMode) {
		case VDShaderFilterConfig::kModeInterp:
			mpInterp = new VDShaderInterpreter;
			mpInterp->LoadSamplers(&mSampler, 1);
			mpInterp->LoadShader(*func, kElements, 1, NULL);
			break;

		case VDShaderFilterConfig::kModeScalarJIT:
		default:
			mpJIT = new VDShaderScalarJIT;
			mpJIT->LoadSamplers(&mSampler, 1);
			mpJIT->LoadShader(*func, kElements, 1, NULL);
			break;
	}
}

void VDShaderFilter::Stop() {
	if (mpJIT) {
		delete mpJIT;
		mpJIT = NULL;
	}

	if (mpInterp) {
		delete mpInterp;
		mpInterp = NULL;
	}

	if (mpTexture) {
		delete[] mpTexture;
		mpTexture = NULL;
	}

	mpEffect = NULL;
}

bool VDShaderFilter::Run() {
	uint32 srcw = fa->src.w;
	uint32 srch = fa->src.h;

	const uint32 *src = (const uint32 *)fa->src.data;
	uint32 *dst = vdptroffset(mpTexture, mTexturePitch) + 1;
	for(uint32 y=0; y<srch; ++y) {
		memcpy(dst, src, srcw*sizeof(uint32));
		dst[-1] = dst[0];
		dst[srcw] = dst[srcw-1];

		vdptrstep(src, fa->src.pitch);
		vdptrstep(dst, mTexturePitch);
	}

	memcpy(mpTexture, vdptroffset(mpTexture, mTexturePitch), (srcw + 2)*sizeof(uint32));
	memcpy(vdptroffset(mpTexture, mTexturePitch*(srch + 1)), vdptroffset(mpTexture, mTexturePitch*srch), (srcw + 2)*sizeof(uint32));

	uint32 dstw = fa->dst.w;
	uint32 dsth = fa->dst.h;
	dst = (uint32 *)fa->dst.data;
	float uv[2];
	float dudx = 1.0f / (float)(int)dstw;
	float dvdy = 1.0f / (float)(int)dsth;
	float u0 = 0.5f * dudx;
	float v0 = 0.5f * dvdy;

	uv[1] = v0;
	VDShaderConstant result;

	if (mpJIT) {
		for(uint32 y=0; y<dsth; ++y) {
			uv[0] = u0;
			for(uint32 x=0; x<dstw; ++x) {
				mpJIT->Execute(result, uv);
				dst[x]	= (VDClampedRoundFixedToUint8Fast(result.v[0]) << 16)
						+ (VDClampedRoundFixedToUint8Fast(result.v[1]) <<  8)
						+  VDClampedRoundFixedToUint8Fast(result.v[2]);
				uv[0] += dudx;
			}
			uv[1] += dvdy;
			vdptrstep(dst, fa->dst.pitch);
		}
	} else {
		for(uint32 y=0; y<dsth; ++y) {
			uv[0] = u0;
			for(uint32 x=0; x<dstw; ++x) {
				mpInterp->Execute(result, uv);
				dst[x]	= (VDClampedRoundFixedToUint8Fast(result.v[0]) << 16)
						+ (VDClampedRoundFixedToUint8Fast(result.v[1]) <<  8)
						+  VDClampedRoundFixedToUint8Fast(result.v[2]);
				uv[0] += dudx;
			}
			uv[1] += dvdy;
			vdptrstep(dst, fa->dst.pitch);
		}
	}
	return true;
}

int VDShaderFilter::Serialize(char *buf, int maxbuf) {
	return 0;
}

int VDShaderFilter::Deserialize(const char *buf, int maxbuf) {
	return 0;
}

void VDShaderFilter::GetSettingString(char *buf, int maxlen) {
	_snprintf(buf, maxlen, " (%s)", mConfig.mFilename);
}

void VDShaderFilter::GetScriptString(char *buf, int maxlen) {
	VDStringA tmp;
	const char *s = mConfig.mFilename;
	while(char c = *s++) {
		if ((unsigned char)((c & 0x7f) - 0x20) >= 0x60) {
			tmp += VDStringA().sprintf("\\x%02x", (unsigned char)c);
			continue;
		}

		if (c == '\\' || c == '"')
			tmp.push_back('\\');
		tmp.push_back(c);
	}
	_snprintf(buf, maxlen, "Config(\"%s\", %d)", tmp.c_str(), mConfig.mExecutionMode);
}

bool VDShaderFilter::Configure(HWND hwnd) {
	VDShaderFilterConfig old(mConfig);

	bool success = VDOpenShaderEditor(hwnd, mConfig, fa->ifp, g_VDFilterAPIVersion >= 11 ? fa->ifp2 : NULL);

	mConfig.mpEffect = NULL;

	if (!success) {
		mConfig = old;
		return false;
	}

	return true;
}

void VDShaderFilter::ScriptConfig(IScriptInterpreter *, void *pfa0, const CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)pfa0;
	VDShaderFilter *pThis = (VDShaderFilter *)fa->filter_data;
	vdstrlcpy(pThis->mConfig.mFilename, *argv[0].asString(), sizeof(pThis->mConfig.mFilename)/sizeof(pThis->mConfig.mFilename[0]));

	unsigned mode = argv[1].asInt();

	if (mode >= (unsigned)VDShaderFilterConfig::kModeCount)
		mode = VDShaderFilterConfig::kModeDefault;
	pThis->mConfig.mExecutionMode = mode;
}

///////////////////////////////////////////////////////////////////////////
VDFilterDefinition<VDShaderFilter> g_customFilterDef("Avery Lee", "vdshader", "Applies a custom pixel shader to incoming video. (Version 1.0)");

void VDRegisterShaderEditor();
void VDUnregisterShaderEditor();

extern "C" __declspec(dllexport) int __cdecl VirtualdubFilterModuleInit2(struct FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	VDRegistryAppKey::setDefaultKey("Software\\Freeware\\VDShader\\");
	VDRegisterShaderEditor();

	ff->addFilter(fm, &g_customFilterDef, sizeof(FilterDefinition));

	g_VDFilterAPIVersion = vdfd_ver;

	vdfd_ver		= VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat		= 9;	// we need this version for copy constructor support

	return 0;
}

extern "C" __declspec(dllexport) void __cdecl VirtualdubFilterModuleDeinit(struct FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(&g_customFilterDef);
	VDUnregisterShaderEditor();
}
