//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <vector>
#include "parser.h"
#include "interp.h"
#include "log.h"

///////////////////////////////////////////////////////////////////////////////
void VDShaderInterpreterSampler::Init(const uint32 *src, ptrdiff_t pitch, uint32 w, uint32 h, bool clampU, bool clampV, bool bilinear) {
	mpSrc		= src;
	mPitch		= pitch;
	mFWidth		= (float)w;
	mFHeight	= (float)h;
	mULimit		= w - 1;
	mVLimit		= h - 1;
	mbClampU	= clampU;
	mbClampV	= clampV;
	mbBilinear	= bilinear;

	mFWidth0 = mFWidth;
	mFHeight0 = mFHeight;
	if (bilinear) {
		mFWidth *= 64.0f;
		mFHeight *= 64.0f;
		mULimit <<= 6;
		mVLimit <<= 6;
	}
}

uint32 VDShaderInterpreterSampler::Sample(float u, float v) const {
	if (!mbClampU)
		u -= floorf(u);

	if (!mbClampV)
		v -= floorf(v);

	u *= mFWidth;
	v *= mFHeight;

	if (mbBilinear) {
		int iu = (int)u;
		int iv = (int)v;

		if (mbClampU && (unsigned)iu > mULimit)
			iu = ~(iu >> 31) & mULimit;
		if (mbClampV && (unsigned)iv > mVLimit)
			iv = ~(iv >> 31) & mVLimit;

		iu -= 32;
		iv -= 32;

		const uint32 *row0 = (const uint32 *)((const char *)mpSrc + mPitch*(iv >> 6)) + (iu >> 6);
		const uint32 *row1 = (const uint32 *)((const char *)row0 + mPitch);
		uint32 px0 = row0[0];
		uint32 px1 = row0[1];
		uint32 px2 = row1[0];
		uint32 px3 = row1[1];

		uint32 px0_rb = px0 & 0xff00ff;
		uint32 px1_rb = px1 & 0xff00ff;
		uint32 px2_rb = px2 & 0xff00ff;
		uint32 px3_rb = px3 & 0xff00ff;
		uint32 px0_ag = px0 & 0xff00ff00;
		uint32 px1_ag = px1 & 0xff00ff00;
		uint32 px2_ag = px2 & 0xff00ff00;
		uint32 px3_ag = px3 & 0xff00ff00;

		sint32 uf = iu & 63;
		sint32 vf = iv & 63;

		uint32 px01_rb = (px0_rb + (((sint32)(px1_rb - px0_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px23_rb = (px2_rb + (((sint32)(px3_rb - px2_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px01_ag = (px0_ag + ((sint32)((px1_ag >> 6) - (px0_ag >> 6))*uf + 0x800080)) & 0xff00ff00;
		uint32 px23_ag = (px2_ag + ((sint32)((px3_ag >> 6) - (px2_ag >> 6))*uf + 0x800080)) & 0xff00ff00;

		uint32 px_rb = (px01_rb + (((sint32)(px23_rb - px01_rb)*vf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px_ag = (px01_ag + ((sint32)((px23_ag >> 6) - (px01_ag >> 6))*vf + 0x800080)) & 0xff00ff00;

		return px_rb + px_ag;
	} else {
		int iu = (int)u;
		int iv = (int)v;

		if (mbClampU && (unsigned)iu > mULimit)
			iu = ~(iu >> 31) & mULimit;
		if (mbClampV && (unsigned)iv > mVLimit)
			iv = ~(iv >> 31) & mVLimit;

		return ((const uint32 *)((const char *)mpSrc + mPitch*iv))[iu];
	}
}

///////////////////////////////////////////////////////////////////////////////
void Disassemble(IVDShaderLogOutput &out, const VDShaderInterpreterTransformedInsn& insn, int tempBase) {
	switch(insn.op) {
		case kVDShaderInterpOp_end:
			out.WriteLogOutput("end");
			break;
		case kVDShaderInterpOp_mov:
			out.WriteLogOutput("mov");
			break;
		case kVDShaderInterpOp_add:
			out.WriteLogOutput("add");
			break;
		case kVDShaderInterpOp_sub:
			out.WriteLogOutput("sub");
			break;
		case kVDShaderInterpOp_mul:
			out.WriteLogOutput("mul");
			break;
		case kVDShaderInterpOp_rsq:
			out.WriteLogOutput("rsq");
			break;
		case kVDShaderInterpOp_rcp:
			out.WriteLogOutput("rcp");
			break;
		case kVDShaderInterpOp_neg:
			out.WriteLogOutput("neg");
			break;
		case kVDShaderInterpOp_texld:
			out.WriteLogOutput("texld");
			break;
		case kVDShaderInterpOp_unpackr:
			out.WriteLogOutput("unpackr");
			break;
		case kVDShaderInterpOp_unpackg:
			out.WriteLogOutput("unpackg");
			break;
		case kVDShaderInterpOp_unpackb:
			out.WriteLogOutput("unpackb");
			break;
		case kVDShaderInterpOp_unpacka:
			out.WriteLogOutput("unpacka");
			break;
		default:
			VDASSERT(false);
	}

	if (insn.dst >= 0) {
		if (insn.dst >= tempBase)
			VDShaderWriteLogOutputF(out, " r%d", insn.dst - tempBase);
		else
			VDShaderWriteLogOutputF(out, " c%d", insn.dst);
	}

	if (insn.src1 >= 0) {
		if (insn.dst >= 0)
			out.WriteLogOutput(",");

		if (insn.src1 >= tempBase)
			VDShaderWriteLogOutputF(out, " r%d", insn.src1 - tempBase);
		else
			VDShaderWriteLogOutputF(out, " c%d", insn.src1);
	}

	if (insn.src2 >= 0) {
		if (insn.src2 >= tempBase)
			VDShaderWriteLogOutputF(out, ", r%d", insn.src2 - tempBase);
		else
			VDShaderWriteLogOutputF(out, ", c%d", insn.src2);
	}

	out.WriteLogOutput("\n");
}

void Disassemble(IVDShaderLogOutput& out, const vdfastvector<VDShaderInterpreterTransformedInsn>& insns, int tempBase) {
	for(vdfastvector<VDShaderInterpreterTransformedInsn>::const_iterator it(insns.begin()), itEnd(insns.end()); it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		Disassemble(out, insn, tempBase);
	}
}

///////////////////////////////////////////////////////////////////////////////
class VDShaderInterpreterRegisterAllocator {
public:
	VDShaderInterpreterRegisterAllocator();
	~VDShaderInterpreterRegisterAllocator();

	void Finalize(int constantCount, int outputBegin, int outputEnd);
	void ExtractInsns(vdfastvector<VDShaderInterpreterTransformedInsn>& insns);

	// get compiled register (slot) count
	int GetSlotCount() const { return mSlotCount; }
	int GetInputBase() const { return mInputBase; }
	int GetTempBase() const { return mTempBase; }
	int GetRegisterMapping(int srcreg) const { return mRegisterAssignments[srcreg]; }

	void mov(int dst, int src);
	void add(int dst, int src1, int src2);
	void sub(int dst, int src1, int src2);
	void mul(int dst, int src1, int src2);
	void rsq(int dst, int src);
	void rcp(int dst, int src);
	void neg(int dst, int src);
	void texld(int dst, int srcu, int srcv);
	void unpackr(int dst, int src);
	void unpackg(int dst, int src);
	void unpackb(int dst, int src);
	void unpacka(int dst, int src);

protected:
	void push_insn(int op, int dst, int src1, int src2);
	void AnalyzeDataFlow(int constantCount, int outputBegin, int outputEnd);
	void AllocateRegisters();

	vdfastvector<int> mRegisterAssignments;

	vdfastvector<VDShaderInterpreterTransformedInsn> mInsns;
	vdfastvector<int> mLastRegRead;

	int mRegCount;
	int mSlotCount;
	int mInputBase;
	int mTempBase;
};

VDShaderInterpreterRegisterAllocator::VDShaderInterpreterRegisterAllocator()
	: mRegCount(0)
	, mSlotCount(0)
	, mInputBase(0)
{
}

VDShaderInterpreterRegisterAllocator::~VDShaderInterpreterRegisterAllocator() {
}

void VDShaderInterpreterRegisterAllocator::Finalize(int constantCount, int outputBegin, int outputEnd) {
	AnalyzeDataFlow(constantCount, outputBegin, outputEnd);
	AllocateRegisters();

	VDShaderInterpreterTransformedInsn sentinel = {0};
	mInsns.push_back(sentinel);
}

void VDShaderInterpreterRegisterAllocator::ExtractInsns(vdfastvector<VDShaderInterpreterTransformedInsn>& insns) {
	insns.swap(mInsns);
}

void VDShaderInterpreterRegisterAllocator::mov(int dst, int src) {
	push_insn(kVDShaderInterpOp_mov, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::add(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_add, dst, src1, src2);
}

void VDShaderInterpreterRegisterAllocator::sub(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_sub, dst, src1, src2);
}

void VDShaderInterpreterRegisterAllocator::mul(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_mul, dst, src1, src2);
}

void VDShaderInterpreterRegisterAllocator::rsq(int dst, int src) {
	push_insn(kVDShaderInterpOp_rsq, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::rcp(int dst, int src) {
	push_insn(kVDShaderInterpOp_rcp, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::neg(int dst, int src) {
	push_insn(kVDShaderInterpOp_neg, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::texld(int dst, int srcu, int srcv) {
	push_insn(kVDShaderInterpOp_texld, dst, srcu, srcv);
}

void VDShaderInterpreterRegisterAllocator::unpackr(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackr, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::unpackg(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackg, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::unpackb(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackb, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::unpacka(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpacka, dst, src, -1);
}

void VDShaderInterpreterRegisterAllocator::push_insn(int op, int dst, int src1, int src2) {
	VDShaderInterpreterTransformedInsn tinsn;
	tinsn.op = op;
	tinsn.dst = dst;
	tinsn.src1 = src1;
	tinsn.src2 = src2;
	mInsns.push_back(tinsn);

	if (dst >= mRegCount)
		mRegCount = dst+1;

	if (src1 >= mRegCount)
		mRegCount = src1+1;

	if (src2 >= mRegCount)
		mRegCount = src2+1;
}

void VDShaderInterpreterRegisterAllocator::AnalyzeDataFlow(int constantCount, int outputBegin, int outputEnd) {
	vdfastvector<int> lastWrite(mRegCount, -1);
	mLastRegRead.resize(mRegCount, -1);

	vdfastvector<bool> isInput(mRegCount, false);

	int numInsns = mInsns.size();

	for(int i=outputBegin; i<outputEnd; ++i) {
		mLastRegRead[i] = numInsns;
	}

	for(int i=numInsns-1; i>=0; --i) {
		VDShaderInterpreterTransformedInsn& insn = mInsns[i];

		if (insn.dst >= 0) {
			if (mLastRegRead[insn.dst] < 0 && (insn.dst < outputBegin || insn.dst >= outputEnd)) {
				insn.op = kVDShaderInterpOp_mov;
				insn.dst = -1;
				insn.src1 = -1;
				insn.src2 = -1;
				continue;
			}

			VDASSERT(lastWrite[insn.dst] < 0);

			lastWrite[insn.dst] = i;
			isInput[insn.dst] = false;
		}

		if (insn.src1 >= 0) {
			if (mLastRegRead[insn.src1] < 0) {
				mLastRegRead[insn.src1] = i;
				isInput[insn.src1] = true;
			}
		}

		if (insn.src2 >= 0) {
			if (mLastRegRead[insn.src2] < 0) {
				mLastRegRead[insn.src2] = i;
				isInput[insn.src2] = true;
			}
		}
	}

	mRegisterAssignments.resize(mRegCount, -1);
	mSlotCount = 0;
	for(int i=0; i<constantCount; ++i) {
		if (isInput[i]) {
			mRegisterAssignments[i] = mSlotCount++;
			mLastRegRead[i] = numInsns;
		}
	}
	mInputBase = mSlotCount;
	for(int i=constantCount; i<mRegCount; ++i) {
		if (isInput[i]) {
			mRegisterAssignments[i] = mSlotCount++;
			mLastRegRead[i] = numInsns;
		}
	}
}

void VDShaderInterpreterRegisterAllocator::AllocateRegisters() {
	vdfastvector<int> slotRefCounts(mSlotCount, 0);
	vdfastvector<int> freeSlots;

	for(int i=0; i<mRegCount; ++i) {
		int slot = mRegisterAssignments[i];

		if (slot >= 0)
			slotRefCounts[slot] += 2;
	}

	mTempBase = mSlotCount;

	int numInsns = mInsns.size();
	int index = 0;
	for(int i=0; i<numInsns; ++index) {
		VDShaderInterpreterTransformedInsn& insn = mInsns[i];
		int r1, r2;

		if (insn.op == kVDShaderInterpOp_mov && insn.dst < 0) {
			mInsns.erase(mInsns.begin() + i);
			--numInsns;
			continue;
		}

		if (insn.src1 >= 0) {
			r1 = mRegisterAssignments[insn.src1];

			VDASSERT(r1 >= 0);

			if (insn.op == kVDShaderInterpOp_mov && insn.dst >= 0)
				++slotRefCounts[r1];

			if (mLastRegRead[insn.src1] == index && insn.src1 != insn.src2) {
				mRegisterAssignments[insn.src1] = -1;

				int rc = --slotRefCounts[r1];
				VDASSERT(rc >= 0);

				if (!rc)
					freeSlots.push_back(r1);
			}

			insn.src1 = r1;
		}

		if (insn.src2 >= 0) {
			r2 = mRegisterAssignments[insn.src2];

			VDASSERT(r2 >= 0);

			if (mLastRegRead[insn.src2] == index) {
				mRegisterAssignments[insn.src2] = -1;

				int rc = --slotRefCounts[r2];
				VDASSERT(rc >= 0);

				if (!rc)
					freeSlots.push_back(r2);
			}

			insn.src2 = r2;
		}

		if (insn.dst >= 0) {
			VDASSERT(mRegisterAssignments[insn.dst] < 0);

			if (insn.op == kVDShaderInterpOp_mov) {
				mRegisterAssignments[insn.dst] = r1;

				--numInsns;
				mInsns.erase(mInsns.begin() + i);
				continue;
			} else {
				int rdst;
				if (!freeSlots.empty()) {
					rdst = freeSlots.back();
					freeSlots.pop_back();
				} else {
					rdst = mSlotCount++;
					slotRefCounts.push_back(0);
				}

				VDASSERT(!slotRefCounts[rdst]);

				slotRefCounts[rdst] = 1;
				mRegisterAssignments[insn.dst] = rdst;
				insn.dst = rdst;
			}
		}

		++i;
	}
}

///////////////////////////////////////////////////////////////////////////////

namespace {
	struct DecodedInsn {
		int op;
		int dst[4];
		int src[16];
	};
}

VDShaderScalarForm::VDShaderScalarForm() {
}

VDShaderScalarForm::~VDShaderScalarForm() {
}

void VDShaderScalarForm::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	VDShaderInterpreterRegisterAllocator regAlloc;

	int numConstants = func.mConstants.size() * 4;

	vdfastvector<int> regAssignments(func.mRegisterCount*4 + 4);
	for(int i=0; i<func.mRegisterCount*4+4; ++i)
		regAssignments[i] = i+numConstants;

	int outputReg = numConstants + func.mRegisterCount*4;
	int nextDstReg = outputReg + 4;
	for(VDShaderFunction::Insns::const_iterator it(func.mInsns.begin()), itEnd(func.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInsn& insn = *it;
		DecodedInsn dinsn;

		for(int i=0; i<4; ++i) {
			const VDShaderInsnSource& src = insn.mSources[i];

			if (!src.mRegister)
				continue;

			int srcRegIndex = src.mRegister & kVDShaderRegIndexMask;
			int srcBase = i*4;

			switch(src.mRegister & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					dinsn.src[srcBase+0] = regAssignments[srcRegIndex*4 + ((src.mSwizzle >> 0) & 3)];
					dinsn.src[srcBase+1] = regAssignments[srcRegIndex*4 + ((src.mSwizzle >> 2) & 3)];
					dinsn.src[srcBase+2] = regAssignments[srcRegIndex*4 + ((src.mSwizzle >> 4) & 3)];
					dinsn.src[srcBase+3] = regAssignments[srcRegIndex*4 + ((src.mSwizzle >> 6) & 3)];
					break;

				case kVDShaderRegC0:
					dinsn.src[srcBase+0] = srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				case kVDShaderRegS0:
					break;

				default:
					VDASSERT(false);
					break;
			}

			if (src.mModifiers & kVDShaderMod_Negate) {
				for(int j=0; j<4; ++j) {
					regAlloc.neg(nextDstReg, dinsn.src[srcBase+j]);
					dinsn.src[srcBase+j] = nextDstReg++;
				}
			}
		}

		int dstReg = insn.mDestRegister;
		int dstRegIndex = dstReg & kVDShaderRegIndexMask;
		int writeMask = insn.mWriteMask;
		bool reallocateDst = false;

		switch(dstReg & kVDShaderRegTypeMask) {
			case kVDShaderRegO0:
				dstRegIndex = func.mRegisterCount * 4;
				break;

			case kVDShaderRegR0:
				dstRegIndex = dstRegIndex*4;
				reallocateDst = true;
				break;

			default:
				VDASSERT(false);
				break;
		}

		for(int i=0; i<4; ++i) {
			if (writeMask & (1 << i)) {
				if (reallocateDst)
					regAssignments[dstRegIndex + i] = nextDstReg++;

				dinsn.dst[i] = regAssignments[dstRegIndex + i];
			} else
				dinsn.dst[i] = -1;
		}

		switch(insn.mOp) {
			case kVDShaderOp_mov:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i] >= 0)
						regAlloc.mov(dinsn.dst[i], dinsn.src[i]);
				}
				break;
			case kVDShaderOp_add:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i] >= 0)
						regAlloc.add(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_sub:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i] >= 0)
						regAlloc.sub(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mul:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i] >= 0)
						regAlloc.mul(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mad:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i] >= 0) {
						int t = nextDstReg++;
						regAlloc.mul(t, dinsn.src[i], dinsn.src[i+4]);
						regAlloc.add(dinsn.dst[i], t, dinsn.src[i+8]);
					}
				}
				break;
			case kVDShaderOp_dp2:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = -1;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i] >= 0) {
							if (t3 < 0) {
								t3 = dinsn.dst[i];

								regAlloc.mul(t, dinsn.src[0], dinsn.src[4]);
								regAlloc.mul(t2, dinsn.src[1], dinsn.src[5]);
								regAlloc.add(t3, t, t2);
							} else
								regAlloc.mov(dinsn.dst[i], t3);
						}
					}
				}
				break;
			case kVDShaderOp_dp3:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = -1;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i] >= 0) {
							if (t5 < 0) {
								t5 = dinsn.dst[i];

								regAlloc.mul(t, dinsn.src[0], dinsn.src[4]);
								regAlloc.mul(t2, dinsn.src[1], dinsn.src[5]);
								regAlloc.add(t3, t, t2);
								regAlloc.mul(t4, dinsn.src[2], dinsn.src[6]);
								regAlloc.add(t5, t3, t4);
							} else
								regAlloc.mov(dinsn.dst[i], t5);
						}
					}
				}
				break;
			case kVDShaderOp_dp4:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = nextDstReg++;
					int t6 = nextDstReg++;
					int t7 = -1;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i] >= 0) {
							if (t7 < 0) {
								t7 = dinsn.dst[i];

								regAlloc.mul(t, dinsn.src[0], dinsn.src[4]);
								regAlloc.mul(t2, dinsn.src[1], dinsn.src[5]);
								regAlloc.add(t3, t, t2);
								regAlloc.mul(t4, dinsn.src[2], dinsn.src[6]);
								regAlloc.add(t5, t3, t4);
								regAlloc.mul(t6, dinsn.src[3], dinsn.src[7]);
								regAlloc.add(t7, t5, t6);
							} else
								regAlloc.mov(dinsn.dst[i], t7);
						}
					}
				}
				break;
			case kVDShaderOp_rsq:
				{
					int t = -1;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i] >= 0) {
							if (t < 0) {
								t = dinsn.dst[i];

								regAlloc.rsq(t, dinsn.src[0]);
							} else
								regAlloc.mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_rcp:
				{
					int t = -1;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i] >= 0) {
							if (t < 0) {
								t = dinsn.dst[i];

								regAlloc.rcp(t, dinsn.src[0]);
							} else
								regAlloc.mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_texld:
				{
					int t = nextDstReg++;
					regAlloc.texld(t, dinsn.src[0], dinsn.src[1]);

					if (dinsn.dst[0] >= 0)
						regAlloc.unpackr(dinsn.dst[0], t);

					if (dinsn.dst[1] >= 0)
						regAlloc.unpackg(dinsn.dst[1], t);

					if (dinsn.dst[2] >= 0)
						regAlloc.unpackb(dinsn.dst[2], t);

					if (dinsn.dst[3] >= 0)
						regAlloc.unpacka(dinsn.dst[3], t);
				}
				break;
		}
	}

	regAlloc.Finalize(numConstants, outputReg, outputReg + 4);

	regAlloc.ExtractInsns(mInsns);

	int tempBase = regAlloc.GetTempBase();

	mConstants.resize(numConstants, 0);

	mSlotCount = regAlloc.GetSlotCount();
	mInputBase = regAlloc.GetInputBase();
	mTempBase = regAlloc.GetTempBase();

	mInputMappings.clear();
	int srcSlots = func.mParameterRegisterCount * 4;;
	for(int i=0; i<srcSlots; ++i) {
		int rsrc = regAlloc.GetRegisterMapping(numConstants + i);

		if (rsrc >= 0) {
			InputMapping imap;
			imap.mSrcOffset = sizeof(float) * i;
			imap.mDstOffset = sizeof(float) * rsrc;
			mInputMappings.push_back(imap);
		}
	}

	// load constants
	for(int i=0; i<numConstants; ++i) {
		int rsrc = regAlloc.GetRegisterMapping(i);

		if (rsrc >= 0) {
			float v = func.mConstants[i >> 2].v[i & 3];
			mConstants[rsrc] = v;
		}
	}

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Generated by VirtualDub Vector-to-Scalar Recompiler\n");
		out->WriteLogOutput("//\n");

		for(int i=0; i<numConstants; ++i)
			VDShaderWriteLogOutputF(*out, "def c%d, %g\n", i, mConstants[i]);

		Disassemble(*out, mInsns, tempBase);
	}

	for(int i=0; i<4; ++i)
		mOutputMappings[i] = regAlloc.GetRegisterMapping(outputReg + i);
}

///////////////////////////////////////////////////////////////////////////////

VDShaderInterpreter::VDShaderInterpreter() {
}

VDShaderInterpreter::~VDShaderInterpreter() {
}

void VDShaderInterpreter::LoadSamplers(const VDShaderInterpreterSampler *samplers, int numSamplers) {
	mpSamplers = samplers;
}

void VDShaderInterpreter::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	mScalarForm.LoadShader(func, pElements, numElements, out);

	mRegisterFile.reserve(mScalarForm.mSlotCount);
	mRegisterFile = mScalarForm.mConstants;
	mRegisterFile.resize(mScalarForm.mSlotCount, 0);
}

void VDShaderInterpreter::Execute(VDShaderConstant& result, const float *args) {
	VDShaderScalarForm::TransformedInsn *ip = &mScalarForm.mInsns[0];
	float *__restrict const r = mRegisterFile.empty() ? NULL : &mRegisterFile[0];

	for(VDShaderScalarForm::InputMappings::const_iterator it(mScalarForm.mInputMappings.begin()), itEnd(mScalarForm.mInputMappings.end()); it!=itEnd; ++it) {
		const VDShaderScalarForm::InputMapping& imap = *it;

		*(float *)((char *)r + imap.mDstOffset) = *(const float *)((const char *)args + imap.mSrcOffset);
	}

	for(;;) {
		int op = ip->op;
		if (!op)
			break;

#define R(index) r[ip->index]

		switch(ip->op) {
			case kVDShaderInterpOp_mov:
				R(dst) = R(src1);
				break;

			case kVDShaderInterpOp_add:
				R(dst) = R(src1) + R(src2);
				break;

			case kVDShaderInterpOp_sub:
				R(dst) = R(src1) - R(src2);
				break;

			case kVDShaderInterpOp_mul:
				R(dst) = R(src1) * R(src2);
				break;

			case kVDShaderInterpOp_rsq:
				R(dst) = 1.0f / sqrtf(fabsf(R(src1)));
				break;

			case kVDShaderInterpOp_rcp:
				R(dst) = 1.0f / R(src1);
				break;

			case kVDShaderInterpOp_neg:
				R(dst) = 1.0f / R(src1);
				break;

			case kVDShaderInterpOp_texld:
				(uint32&)R(dst) = mpSamplers[0].Sample(R(src1), R(src2));
				break;

			case kVDShaderInterpOp_unpackr:
				R(dst) = (float)(((uint32&)R(src1) >> 16) & 0xff) * (1.0f / 255.0f);
				break;

			case kVDShaderInterpOp_unpackg:
				R(dst) = (float)(((uint32&)R(src1) >>  8) & 0xff) * (1.0f / 255.0f);
				break;

			case kVDShaderInterpOp_unpackb:
				R(dst) = (float)(((uint32&)R(src1)      ) & 0xff) * (1.0f / 255.0f);
				break;

			case kVDShaderInterpOp_unpacka:
				R(dst) = (float)(((uint32&)R(src1) >> 24) & 0xff) * (1.0f / 255.0f);
				break;

		}

#undef R

		++ip;
	}

	result.v[0] = r[mScalarForm.mOutputMappings[0]];
	result.v[1] = r[mScalarForm.mOutputMappings[1]];
	result.v[2] = r[mScalarForm.mOutputMappings[2]];
	result.v[3] = r[mScalarForm.mOutputMappings[3]];
}

