//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "parser.h"
#include "intrinsics.h"
#include "vecarith.h"

using namespace nsVDShaderParser;
using namespace nsVDShaderArith;

bool VDShaderCompileIntrinsic_dot(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& vB = args[1];
	VDShaderParserVariable& vA = args[0];

	int rA = 0;
	
	if (vA.mbIsConst != vB.mbIsConst) {
		func.PromoteToConstantRegisters(vA);
		func.PromoteToConstantRegisters(vB);
	}

	if (vA.mbIsConst) {
		switch(vA.mType) {
			case kTypeFloat2:
				vA.m[0] = dp2(vA.m[0], vB.m[0]);
				break;
			case kTypeFloat3:
				vA.m[0] = dp3(vA.m[0], vB.m[0]);
				break;
			case kTypeFloat4:
				vA.m[0] = dp4(vA.m[0], vB.m[0]);
				break;
		}
		vA.m[0].v[1] = 0.0f;
		vA.m[0].v[2] = 0.0f;
		vA.m[0].v[3] = 0.0f;
	} else {
		rA = func.AllocateRegisters(1);

		switch(vA.mType) {
			case kTypeFloat2:
				func.dp2(rA, 0x1, vA.mRegister, vB.mRegister);
				break;
			case kTypeFloat3:
				func.dp3(rA, 0x1, vA.mRegister, vB.mRegister);
				break;
			case kTypeFloat4:
				func.dp4(rA, 0x1, vA.mRegister, vB.mRegister);
				break;
		}
	}

	vA.mNameToken = -1;
	vA.mType = kTypeFloat;
	vA.mRegister = rA;
	return true;
}

bool VDShaderCompileIntrinsic_length(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];

	if (v.mbIsConst) {
		switch(v.mType) {
			case kTypeFloat2:
				v.m[0].v[0] = sqrtf(dp2(v.m[0], v.m[0]).v[0]);
				break;
			case kTypeFloat3:
				v.m[0].v[0] = sqrtf(dp3(v.m[0], v.m[0]).v[0]);
				break;
			case kTypeFloat4:
				v.m[0].v[0] = sqrtf(dp4(v.m[0], v.m[0]).v[0]);
				break;
		}
		v.m[0].v[1] = 0.0f;
		v.m[0].v[2] = 0.0f;
		v.m[0].v[3] = 0.0f;
	} else {
		int rA = func.AllocateRegisters(3);
		int rB = rA + 1;
		int rC = rB + 1;

		switch(v.mType) {
			case kTypeFloat2:
				func.dp2(rA, 0x1, v.mRegister, v.mRegister);
				break;
			case kTypeFloat3:
				func.dp3(rA, 0x1, v.mRegister, v.mRegister);
				break;
			case kTypeFloat4:
				func.dp4(rA, 0x1, v.mRegister, v.mRegister);
				break;
		}

		func.rsq(rB, 0x1, rA);
		func.rcp(rC, 0x1, rB);
		v.mRegister = rC;
	}

	v.mNameToken = -1;
	v.mType = kTypeFloat;
	return true;
}

bool VDShaderCompileIntrinsic_normalize(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];

	if (v.mbIsConst) {
		float sum = 0.0f;

		switch(v.mType) {
			case kTypeFloat4:
				sum = dp2(v.m[0], v.m[0]).v[0];
				break;
			case kTypeFloat3:
				sum = dp3(v.m[0], v.m[0]).v[0];
				break;
			case kTypeFloat2:
				sum = dp4(v.m[0], v.m[0]).v[0];
				break;
		}

		if (fabsf(sum) < 1e-30f)
			return lexer.SetError("Attempted normalize() on zero-magnitude vector");

		float invscale = 1.0f / sqrtf(sum);

		v.m[0].v[0] *= invscale;
		v.m[0].v[1] *= invscale;
		v.m[0].v[2] *= invscale;
		v.m[0].v[3] *= invscale;
	} else {
		int rA = func.AllocateRegisters(3);
		int rB = rA + 1;
		int rC = rB + 1;

		unsigned char finalMask;
		switch(v.mType) {
			case kTypeFloat2:
				func.dp2(rA, 0x1, v.mRegister, v.mRegister);
				finalMask = 0x3;
				break;
			case kTypeFloat3:
				func.dp3(rA, 0x1, v.mRegister, v.mRegister);
				finalMask = 0x7;
				break;
			case kTypeFloat4:
				func.dp4(rA, 0x1, v.mRegister, v.mRegister);
				finalMask = 0xf;
				break;
		}

		func.rsq(rB, 0x1, rA);
		func.mul_swizzle(rC, finalMask, v.mRegister, 0xE4, rB, 0);
		v.mRegister = rC;
	}

	v.mNameToken = -1;
	return true;
}

bool VDShaderCompileIntrinsic_mul_v2_m2x2(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];
	VDShaderParserVariable& m = args[1];

	if (v.mbIsConst != m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];

		v.m[0].v[0] = m.m[0].v[0]*x + m.m[1].v[0]*y;
		v.m[0].v[1] = m.m[0].v[1]*x + m.m[1].v[1]*y;
	} else {
		int rA = func.AllocateRegisters(2);

		func.mul_swizzle(rA+0, 0x3, v.mRegister, 0x00, m.mRegister + 0, 0xE4);
		func.mad_swizzle(rA+1, 0x3, v.mRegister, 0x55, m.mRegister + 1, 0xE4, rA+0, 0xE4);
		v.mRegister = rA + 1;
	}

	v.mNameToken = -1;
	return true;
}

bool VDShaderCompileIntrinsic_mul_v3_m3x3(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];
	VDShaderParserVariable& m = args[1];

	if (v.mbIsConst != m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];
		float z = v.m[0].v[2];

		v.m[0].v[0] = m.m[0].v[0]*x + m.m[1].v[0]*y + m.m[2].v[0]*z;
		v.m[0].v[1] = m.m[0].v[1]*x + m.m[1].v[1]*y + m.m[2].v[1]*z;
		v.m[0].v[2] = m.m[0].v[2]*x + m.m[1].v[2]*y + m.m[2].v[2]*z;
	} else {
		int rA = func.AllocateRegisters(3);

		func.mul_swizzle(rA+0, 0x7, v.mRegister, 0x00, m.mRegister + 0, 0xE4);
		func.mad_swizzle(rA+1, 0x7, v.mRegister, 0x55, m.mRegister + 1, 0xE4, rA+0, 0xE4);
		func.mad_swizzle(rA+2, 0x7, v.mRegister, 0xAA, m.mRegister + 2, 0xE4, rA+1, 0xE4);
		v.mRegister = rA + 2;
	}

	v.mNameToken = -1;
	return true;
}

bool VDShaderCompileIntrinsic_tex2D(VDShaderLexer& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& samp = args[0];
	VDShaderParserVariable& uv = args[1];

	func.PromoteToConstantRegisters(uv);

	int rA = func.AllocateRegisters(1);

	func.texld(rA, 0xF, uv.mRegister, 0xE4, samp.mRegister);

	samp.mNameToken = -1;
	samp.mType = kTypeFloat4;
	samp.mRegister = rA;
	return true;
}

namespace {
	template<VDShaderParserType t1> struct VDShaderParserArgTypes0 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1> const VDShaderParserType VDShaderParserArgTypes0<t1>::kTypeArray[] = { t1, kTypeInvalid };
	template<VDShaderParserType t1, VDShaderParserType t2> struct VDShaderParserArgTypes1 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1, VDShaderParserType t2> const VDShaderParserType VDShaderParserArgTypes1<t1, t2>::kTypeArray[] = { t1, t2, kTypeInvalid };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3> struct VDShaderParserArgTypes2 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3> const VDShaderParserType VDShaderParserArgTypes2<t1, t2, t3>::kTypeArray[] = { t1, t2, t3, kTypeInvalid };
}

#define ARGS_0(rv) VDShaderParserArgTypes0<kType##rv>::kTypeArray
#define ARGS_1(rv, t1) VDShaderParserArgTypes1<kType##rv, kType##t1>::kTypeArray
#define ARGS_2(rv, t1, t2) VDShaderParserArgTypes2<kType##rv, kType##t1, kType##t2>::kTypeArray

const VDShaderIntrinsicEntry kVDShaderIntrinsics[]={
	{ "dot",		VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float4, Float4) },
	{ "length",		VDShaderCompileIntrinsic_length, ARGS_1(Float, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_length, ARGS_1(Float, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_length, ARGS_1(Float, Float4) },
	{ "normalize",	VDShaderCompileIntrinsic_normalize, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_normalize, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_normalize, ARGS_1(Float4, Float4) },
	{ "mul",		VDShaderCompileIntrinsic_mul_v2_m2x2, ARGS_2(Float2, Float2, Float2x2) },
	{ NULL,			VDShaderCompileIntrinsic_mul_v3_m3x3, ARGS_2(Float3, Float3, Float3x3) },
	{ "tex2D",		VDShaderCompileIntrinsic_tex2D, ARGS_2(Float4, Sampler, Float2) },
	{ NULL, NULL, NULL }
};

const VDShaderIntrinsicEntry *VDShaderParserGetIntrinsicByName(const char *s) {
	const VDShaderIntrinsicEntry *p = kVDShaderIntrinsics;

	while(p->mpCompiler) {
		if (p->mpName && !strcmp(s, p->mpName))
			return p;

		++p;
	}

	return NULL;
}

bool VDShaderParserCompileIntrinsic(VDShaderLexer& lexer, VDShaderParserFunction& func, const VDShaderIntrinsicEntry *pBaseEntry, VDShaderParserVariable *args, int numArgs) {
	const VDShaderIntrinsicEntry *pEntry = pBaseEntry;
	do {
		const VDShaderParserType *pTypeList = pEntry->mTypes + 1;

		int arg = 0;
		for(; arg<numArgs; ++arg) {
			if (pTypeList[arg] != args[arg].mType)
				break;
		}

		if (arg == numArgs && !pTypeList[numArgs])
			return pEntry->mpCompiler(lexer, func, args, numArgs);

		++pEntry;
	} while(pEntry->mpCompiler && !pEntry->mpName);

	return lexer.SetErrorF("No form of intrinsic '%s' matched the given argument list", pBaseEntry->mpName);
}
