//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "jit_scalar.h"
#include "jit.h"
#include "x86disasm.h"
#include "log.h"
#include <xmmintrin.h>

extern const unsigned char g_VDDisasmX86[];

VDShaderScalarJIT::VDShaderScalarJIT()
	: mpShader(NULL)
{
}

VDShaderScalarJIT::~VDShaderScalarJIT() {
	if (mpShader) {
		mAllocator.Free(mpShader, mShaderLen);
		mpShader = NULL;
	}
}

void VDShaderScalarJIT::LoadSamplers(const VDShaderInterpreterSampler *samplers, int numSamplers) {
	mpSamplers = samplers;
}

namespace {
	uint32 SampleTexturePoint(VDShaderInterpreterSampler& samp, float u, float v) {
		if (!samp.mbClampU)
			u -= floorf(u);

		if (!samp.mbClampV)
			v -= floorf(v);

		u *= samp.mFWidth;
		v *= samp.mFHeight;

		int iu = (int)u;
		int iv = (int)v;

		if (samp.mbClampU && (unsigned)iu > samp.mULimit)
			iu = ~(iu >> 31) & samp.mULimit;
		if (samp.mbClampV && (unsigned)iv > samp.mVLimit)
			iv = ~(iv >> 31) & samp.mVLimit;

		return ((const uint32 *)((const char *)samp.mpSrc + samp.mPitch*iv))[iu];
	}

	uint32 SampleTextureBilinear(VDShaderInterpreterSampler& samp, float u, float v) {
		if (!samp.mbClampU)
			u -= floorf(u);

		if (!samp.mbClampV)
			v -= floorf(v);

		u *= samp.mFWidth;
		v *= samp.mFHeight;

		int iu = (int)u;
		int iv = (int)v;

		if (samp.mbClampU && (unsigned)iu > samp.mULimit)
			iu = ~(iu >> 31) & samp.mULimit;
		if (samp.mbClampV && (unsigned)iv > samp.mVLimit)
			iv = ~(iv >> 31) & samp.mVLimit;

		iu -= 32;
		iv -= 32;

		const uint32 *row0 = (const uint32 *)((const char *)samp.mpSrc + samp.mPitch*(iv >> 6)) + (iu >> 6);
		const uint32 *row1 = (const uint32 *)((const char *)row0 + samp.mPitch);
		uint32 px0 = row0[0];
		uint32 px1 = row0[1];
		uint32 px2 = row1[0];
		uint32 px3 = row1[1];

		uint32 px0_rb = px0 & 0xff00ff;
		uint32 px1_rb = px1 & 0xff00ff;
		uint32 px2_rb = px2 & 0xff00ff;
		uint32 px3_rb = px3 & 0xff00ff;
		uint32 px0_ag = px0 & 0xff00ff00;
		uint32 px1_ag = px1 & 0xff00ff00;
		uint32 px2_ag = px2 & 0xff00ff00;
		uint32 px3_ag = px3 & 0xff00ff00;

		sint32 uf = iu & 63;
		sint32 vf = iv & 63;

		uint32 px01_rb = (px0_rb + (((sint32)(px1_rb - px0_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px23_rb = (px2_rb + (((sint32)(px3_rb - px2_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px01_ag = (px0_ag + ((sint32)((px1_ag >> 6) - (px0_ag >> 6))*uf + 0x800080)) & 0xff00ff00;
		uint32 px23_ag = (px2_ag + ((sint32)((px3_ag >> 6) - (px2_ag >> 6))*uf + 0x800080)) & 0xff00ff00;

		uint32 px_rb = (px01_rb + (((sint32)(px23_rb - px01_rb)*vf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px_ag = (px01_ag + ((sint32)((px23_ag >> 6) - (px01_ag >> 6))*vf + 0x800080)) & 0xff00ff00;

		return px_rb + px_ag;
	}
}

void VDShaderScalarJIT::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	if (mpShader) {
		mAllocator.Free(mpShader, mShaderLen);
		mpShader = NULL;
	}

	// compile shader to scalar form
	mScalarForm.LoadShader(func, pElements, numElements, out);

	using namespace x86;
	DynamicContext ctx;

	BasicBlock *bbent = ctx.GetEntryBlock();
	BasicBlock *bbexit = ctx.GetExitBlock();

	int numConstants = mScalarForm.mInputBase;
	const float *pConstants = mScalarForm.mConstants.empty() ? NULL : &mScalarForm.mConstants[0];

	int numTemps = mScalarForm.mSlotCount - mScalarForm.mInputBase;

	bbent->sub(esp, numTemps * 4 + 16);

	static const float kDiv255 = 1.0f / 255.0f;

	bbent->mov(ebx, dword_ptr[esp + numTemps*4 + 40]);
	for(VDShaderScalarForm::InputMappings::const_iterator it(mScalarForm.mInputMappings.begin()), itEnd(mScalarForm.mInputMappings.end()); it!=itEnd; ++it) {
		const VDShaderScalarForm::InputMapping& imap = *it;

		bbent->mov(eax, dword_ptr[ebx + imap.mSrcOffset]);
		bbent->mov(dword_ptr[esp + 16 + imap.mDstOffset - 4*numConstants], eax);
	}

	struct Local {
		Local(BasicBlock *bb, const float *pc, int nc) : mpBlock(bb), mpConstants(pc), mNumConstants(nc) {
		}

		const dword_ref_t GetVariableSlot(int reg, int esp_offset = 0) {
			using namespace x86;
			return reg < mNumConstants ? dword_ptr[mpConstants + reg] : dword_ptr[esp + (reg - mNumConstants)*4 + 16 + esp_offset];
		}

		BasicBlock *mpBlock;
		const float *mpConstants;
		int mNumConstants;
	} local(bbent, pConstants, numConstants);

	VDShaderScalarForm::Insns::const_iterator it(mScalarForm.mInsns.begin()), itEnd(mScalarForm.mInsns.end());
	for(; it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;
		if (!insn.op)
			break;

		switch(insn.op) {
			case kVDShaderInterpOp_texld:
				bbent->push(local.GetVariableSlot(insn.src2, 0));
				bbent->push(local.GetVariableSlot(insn.src1, 4));

				bbent->push((sint32)mpSamplers);

				if (mpSamplers[0].mbBilinear)
					bbent->call(ctx.CreateAbsTarget(SampleTextureBilinear));
				else
					bbent->call(ctx.CreateAbsTarget(SampleTexturePoint));
				bbent->add(esp, 12);
				bbent->mov(dword_ptr[esp + (insn.dst - numConstants)*4 + 16], eax);
				continue;
			case kVDShaderInterpOp_unpackr:
			case kVDShaderInterpOp_unpackg:
			case kVDShaderInterpOp_unpackb:
			case kVDShaderInterpOp_unpacka:
				if (insn.src1 >= numConstants)
					bbent->movzx(eax, byte_ptr[esp + (insn.src1 - numConstants)*4 + (insn.op - kVDShaderInterpOp_unpackb) + 16]);
				else
					bbent->movzx(eax, byte_ptr[(char *)(pConstants + insn.src1) + (insn.op - kVDShaderInterpOp_unpackb)]);
				bbent->mov(dword_ptr [esp], eax);
				bbent->fild(dword_ptr [esp]);
				bbent->fld(dword_ptr [&kDiv255]);
				bbent->fmulp();
				bbent->fstp(local.GetVariableSlot(insn.dst));
				continue;
		}

		if (insn.src1 >= 0)
			bbent->fld(local.GetVariableSlot(insn.src1));

		if (insn.src2 >= 0)
			bbent->fld(local.GetVariableSlot(insn.src2));

		switch(insn.op) {
			case kVDShaderInterpOp_mov:
				break;
			case kVDShaderInterpOp_add:
				bbent->faddp();
				break;
			case kVDShaderInterpOp_sub:
				bbent->fsubp();
				break;
			case kVDShaderInterpOp_mul:
				bbent->fmulp();
				break;
			case kVDShaderInterpOp_rsq:
				bbent->fsqrt();
				bbent->fld1();
				bbent->fdivrp();
				break;
			case kVDShaderInterpOp_rcp:
				bbent->fld1();
				bbent->fdivrp();
				break;
			case kVDShaderInterpOp_neg:
				bbent->fchs();
				break;
		}

		if (insn.dst >= 0)
			bbent->fstp(local.GetVariableSlot(insn.dst));
	}

	bbent->mov(ecx, dword_ptr[esp + numTemps * 4 + 36]);

	for(int i=0; i<4; ++i) {
		int r = mScalarForm.mOutputMappings[i];
		if (r >= numConstants)
			bbent->mov(eax, dword_ptr[esp + (r - numConstants)*4 + 16]);
		else
			bbent->mov(eax, dword_ptr[pConstants + r]);

		bbent->mov(dword_ptr [ecx + i*4], eax);
	}

	bbent->add(esp, numTemps * 4 + 16);
	bbent->jmp(bbexit);

	ctx.Generate();

	uint32 codelen = ctx.GetCodeLen();
	mpShader = (Shader)mAllocator.Allocate(codelen);
	if (mpShader) {
		ctx.Copy(mpShader);
		mAllocator.EndUpdate(mpShader, codelen);

		if (out) {
			out->WriteLogOutput("//\n");
			out->WriteLogOutput("// Generated by VirtualDub Scalar x86 JIT Compiler\n");
			out->WriteLogOutput("//\n");

			VDDisassemblyContext ctx;
			ctx.Init(g_VDDisasmX86);

			const unsigned char *sbase = (const unsigned char *)mpShader;
			const unsigned char *s = sbase;
			const unsigned char *send = sbase + codelen;

			while(s != send) {
				int count;
				const char *text = ctx.Disassemble(s, send - s, count);

				VDShaderWriteLogOutputF(*out, "%4x: %s\n", s - sbase, text);
				s += count;
			}
		}
	}
}
