//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "parser.h"
#include "disasm.h"
#include "interp.h"
#include "intrinsics.h"
#include "vecarith.h"
#include <vd2/system/vdalloc.h>

using namespace nsVDShaderTokens;
using namespace nsVDShaderParser;
using namespace nsVDShaderArith;

namespace {
	int GetBinaryResultType(int srcType, int dstType) {
		static const int kTypeLookup[4][4]={
			kTypeFloat, kTypeFloat2, kTypeFloat3, kTypeFloat4,
			kTypeInvalid, kTypeFloat2x2, kTypeFloat2x3, kTypeFloat2x4,
			kTypeInvalid, kTypeFloat3x2, kTypeFloat3x3, kTypeFloat3x4,
			kTypeInvalid, kTypeFloat4x2, kTypeFloat4x3, kTypeFloat4x4,
		};

		int srcRows = kRowCountForType[srcType];
		int srcCols = kColCountForType[srcType];
		int dstRows = kRowCountForType[dstType];
		int dstCols = kColCountForType[dstType];

		if ((!srcRows && !srcCols) || (!dstRows && !dstCols))
			return 0;

		int maxRows = srcRows > dstRows ? srcRows : dstRows;
		int maxCols = srcCols > dstCols ? srcCols : dstCols;

		return kTypeLookup[maxRows-1][maxCols-1];
	}
}

const char *VDShaderGetTypeName(int type) {
	switch(type) {
	case kTypeVoid:		return "void";
	case kTypeVertexShader:	return "vertexshader";
	case kTypePixelShader:	return "pixelshader";
	case kTypeFloat:	return "float";
	case kTypeFloat2:	return "float2";
	case kTypeFloat3:	return "float3";
	case kTypeFloat4:	return "float4";
	case kTypeFloat2x2:	return "float2x2";
	case kTypeFloat2x3:	return "float2x3";
	case kTypeFloat2x4:	return "float2x4";
	case kTypeFloat3x2:	return "float3x2";
	case kTypeFloat3x3:	return "float3x3";
	case kTypeFloat3x4:	return "float3x4";
	case kTypeFloat4x2:	return "float4x2";
	case kTypeFloat4x3:	return "float4x3";
	case kTypeFloat4x4:	return "float4x4";
	case kTypeSampler:	return "sampler";
	default:			return "<invalid type>";
	}
}

///////////////////////////////////////////////////////////////////////////////
VDShaderFunction::VDShaderFunction() 
	: mRegisterCount(0)
	, mSamplerCount(0)
{
}

VDShaderFunction::~VDShaderFunction() {
}

///////////////////////////////////////////////////////////////////////////////
VDShaderParserFunction::VDShaderParserFunction() {
}

VDShaderParserFunction::~VDShaderParserFunction() {
}

int VDShaderParserFunction::AllocateRegisters(int rows) {
	int base = mRegisterCount;

	mRegisterCount += rows;

	return base + kVDShaderRegR0;
}

int VDShaderParserFunction::AddConstantZero() {
	VDShaderConstant zero = {0};
	return AddConstant(zero);
}

int VDShaderParserFunction::AddConstant(const VDShaderConstant& c) {
	size_t h = VDShaderConstantHash()(c);

	std::pair<ConstantMap::iterator, ConstantMap::iterator> r(mConstantMap.equal_range(h));
	for(; r.first != r.second; ++r.first) {
		const VDShaderConstant& c2 = mConstants[r.first->second];

		if (c == c2)
			return kVDShaderRegC0 + r.first->second;
	}

	int idx = (int)mConstants.size();
	mConstantMap.insert(ConstantMap::value_type(h, idx));

	mConstants.push_back(c);
	return kVDShaderRegC0 + idx;
}

int VDShaderParserFunction::PromoteToConstantRegisters(VDShaderParserVariable& var) {
	if (var.mbIsConst) {
		int rows = kRowCountForType[var.mType];

		int c = kVDShaderRegC0 + (int)mConstants.size();

		mConstants.insert(mConstants.end(), &var.m[0], &var.m[rows]);

		var.mbIsConst = false;
		var.mRegister = c;
	}
	return var.mRegister;
}

int VDShaderParserFunction::AllocateSampler(int globalSamplerIndex) {
	VDShaderFunction::SamplerBindings::const_iterator it(std::find(mSamplerBindings.begin(), mSamplerBindings.end(), globalSamplerIndex));
	int r = kVDShaderRegS0 + (it - mSamplerBindings.begin());
	if (it == mSamplerBindings.end()) {
		mSamplerBindings.push_back(globalSamplerIndex);
		++mSamplerCount;
	}

	VDASSERT(mSamplerBindings.size() == mSamplerCount);

	return r;
}

void VDShaderParserFunction::mov(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_mov, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::mov_swizzle(int dst, unsigned char mask, int src1, unsigned char swizzle) {
	insn(kVDShaderOp_mov, mask, dst, src1, swizzle, 0, 0xE4);
}

void VDShaderParserFunction::mov_mods(int dst, unsigned char mask, int src1, unsigned char swizzle, unsigned char mods) {
	insn_mod(kVDShaderOp_mov, mask, dst, src1, swizzle, mods);
}

void VDShaderParserFunction::add(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_add, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::sub(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_sub, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::mul(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_mul, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::mul_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle) {
	insn(kVDShaderOp_mul, mask, dst, src1, src1swizzle, src2, src2swizzle);
}

void VDShaderParserFunction::mad_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle) {
	insn(kVDShaderOp_mad, mask, dst, src1, src1swizzle, src2, src2swizzle, src3, src3swizzle);
}

void VDShaderParserFunction::dp2(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_dp2, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::dp3(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_dp3, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::dp4(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_dp4, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::rsq(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_rsq, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::rcp(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_rcp, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::rcp(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_rcp, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::texld(int dst, unsigned char mask, int src1, unsigned char src1component, int src2) {
	insn(kVDShaderOp_texld, mask, dst, src1, src1component, src2, 0xE4);
}

void VDShaderParserFunction::insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mod) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = src1mod;
	insn.mSources[1].mRegister = 0;
	insn.mSources[1].mSwizzle = 0;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = 0;
	insn.mSources[2].mSwizzle = 0;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = 0;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = 0;
	insn.mSources[2].mSwizzle = 0;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = 0;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = src3;
	insn.mSources[2].mSwizzle = src3swizzle;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

namespace {
	struct RegisterComponentSource {
		int mReg;
		int mComp;
		unsigned char mMods;
	};

	struct RegisterInfo {
		RegisterComponentSource mComponents[4];
		unsigned char mReferenceMask;
		unsigned char mWriteMask;

		RegisterInfo()
			: mReferenceMask(0)
			, mWriteMask(0)
		{
			for(int i=0; i<4; ++i) {
				mComponents[i].mReg = 0;
				mComponents[i].mComp = 0;
				mComponents[i].mMods = 0;
			}
		}
	};
}

bool VDShaderParserFunction::Inline(VDShaderLexer& lexer, const VDShaderParserFunction& calledFunc, const VDShaderParserVariable *arguments, VDShaderVariable& returnValue) {
	int childBase = AllocateRegisters(calledFunc.mRegisterCount);
	int outputBase = AllocateRegisters(kRowCountForType[calledFunc.mReturnType]);

	std::vector<int> samplerRemap(calledFunc.mSamplerCount, -1);
	for(int i=0; i<calledFunc.mSamplerCount; ++i) {
		int gsindex = calledFunc.mSamplerBindings[i];

		if (gsindex >= 0)
			samplerRemap[i] = AllocateSampler(gsindex);
	}

	int argCount = calledFunc.mParameters.size();
	int argReg = childBase;
	for(int arg=0; arg<argCount; ++arg) {
		const VDShaderParserVariable& argv = arguments[arg];
		int argType = argv.mType;

		if (argType == kTypeSampler) {
			int srcReg = arguments[arg].mRegister;

			samplerRemap[calledFunc.mParameters[arg].mRegister - kVDShaderRegS0] = srcReg;
		} else {
			int rows = kRowCountForType[argType];
			int mask = (1 << kColCountForType[argType]) - 1;

			if (argv.mbIsConst) {
				for(int row=0; row<rows; ++row)
					mov(argReg++, mask, AddConstant(argv.m[row]));
			} else {
				int srcReg = argv.mRegister;
				for(int row=0; row<rows; ++row)
					mov(argReg++, mask, srcReg);
			}
		}
	}

	// copy constants
	int constantCount = calledFunc.mConstants.size();
	std::vector<int> constantRemap(constantCount);
	for(int i=0; i<constantCount; ++i)
		constantRemap[i] = AddConstant(calledFunc.mConstants[i]);

	// inline function body
	for(Insns::const_iterator it(calledFunc.mInsns.begin()), itEnd(calledFunc.mInsns.end()); it!=itEnd; ++it) {
		VDShaderInsn insn = *it;

		// rename destination register
		switch(insn.mDestRegister & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				insn.mDestRegister += childBase - kVDShaderRegR0;
				break;
			case kVDShaderRegO0:
				insn.mDestRegister = outputBase + (insn.mDestRegister & kVDShaderRegIndexMask);
				break;
		}

		// rename source 
		for(int i=0; i<4; ++i) {
			VDShaderInsnSource& src = insn.mSources[i];
			switch(src.mRegister & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				src.mRegister += childBase - kVDShaderRegR0;
				break;
			case kVDShaderRegC0:
				src.mRegister = constantRemap[src.mRegister & kVDShaderRegIndexMask];
				break;
			case kVDShaderRegS0:
				src.mRegister = samplerRemap[src.mRegister & kVDShaderRegIndexMask];
				break;
			}
		}

		mInsns.push_back(insn);
	}

	returnValue.mNameToken = 0;
	returnValue.mSemanticToken = 0;
	returnValue.mRegister = outputBase;
	returnValue.mType = calledFunc.mReturnType;
	return true;
}

bool VDShaderParserFunction::Optimize(VDShaderLexer& lexer, const VDShaderNameTable& nameTable) {
	vdfastvector<RegisterInfo> regInfos(mRegisterCount);
	vdfastvector<int> tempRemap;
	vdfastvector<int> constantRemap;
	vdfastvector<VDShaderConstant> newConstants;

	vdfastvector<VDShaderInsn> newinsns;

	bool eliminatedAnything;
	do {
		eliminatedAnything = false;

		// perform live register analysis
		for(vdfastvector<RegisterInfo>::iterator it(regInfos.begin()), itEnd(regInfos.end()); it!=itEnd; ++it) {
			RegisterInfo& regInfo = *it;

			regInfo.mReferenceMask = 0;
			regInfo.mWriteMask = 0;
			regInfo.mComponents[0].mComp = 0;
			regInfo.mComponents[0].mReg = 0;
			regInfo.mComponents[0].mMods = 0;
			regInfo.mComponents[1].mComp = 0;
			regInfo.mComponents[1].mReg = 0;
			regInfo.mComponents[1].mMods = 0;
			regInfo.mComponents[2].mComp = 0;
			regInfo.mComponents[2].mReg = 0;
			regInfo.mComponents[2].mMods = 0;
			regInfo.mComponents[3].mComp = 0;
			regInfo.mComponents[3].mReg = 0;
			regInfo.mComponents[3].mMods = 0;
		}

		// copy propagation / register packing
		constantRemap.clear();
		constantRemap.resize(mConstants.size(), -1);
		newConstants.clear();

		tempRemap.clear();
		tempRemap.resize(mRegisterCount, -1);
		for(int i=0; i<mParameterRegisterCount; ++i)
			tempRemap[i] = i;
		int newRegisterCount = mParameterRegisterCount;

		for(Insns::iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
			VDShaderInsn& insn = *it;

			unsigned srcmask = kInsnComponentMasks[insn.mOp];

			switch(insn.mOp) {
				case kVDShaderOp_dp2:
				case kVDShaderOp_dp3:
				case kVDShaderOp_dp4:
				case kVDShaderOp_texld:
					break;
				default:
					{
						unsigned insnmask = insn.mWriteMask * 0x11111111;

						srcmask &= insnmask;
					}
					break;
			}

			for(int i=0; i<4; ++i) {
				VDShaderInsnSource& src = insn.mSources[i];

				if (!src.mRegister)
					break;

				if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegC0) {
					int ci = src.mRegister & kVDShaderRegIndexMask;
					int nci = constantRemap[ci];

					if (nci < 0) {
						nci = newConstants.size();
						constantRemap[ci] = nci;
						newConstants.push_back(mConstants[ci]);
					}

					src.mRegister = kVDShaderRegC0 + nci;
				} else if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
					// check for valid register
					int rindex = src.mRegister - kVDShaderRegR0;
					if (rindex >= mRegisterCount)
						return lexer.SetError("Internal compiler error: Invalid temp register as source");

					// remap temp register
					rindex = tempRemap[rindex];
					if (rindex < 0) {
						// We may get a mov instruction that references some uninitialized components
						// for a write merge.
						if (insn.mOp == kVDShaderOp_mov) {
							insn.mOp = kVDShaderOp_nop;
							goto nop_insn;
						} else {
							lexer.SetError("Internal compiler error: Uninitialized temp register used");
							VDShaderDisassembleFunction(*lexer.GetLogOutput(), *this, nameTable);
							return false;
						}
					}
					src.mRegister = kVDShaderRegR0 + rindex;

					// attempt to remap via copy propagation
					do {
						const RegisterInfo& srcRegInfo = regInfos[src.mRegister & kVDShaderRegIndexMask];
						int upstreamReg = 0;
						unsigned char swizzle = 0;
						unsigned char modifiers = 0;

						for(int j=0; j<4; ++j) {
							swizzle >>= 2;

							if (srcmask & (1 << (i*4 + j))) {
								int refComp = (src.mSwizzle >> (j+j)) & 3;
								int refReg = srcRegInfo.mComponents[refComp].mReg;
								unsigned char refMods = srcRegInfo.mComponents[refComp].mMods;

								if (!refReg) {
									upstreamReg = 0;
									break;
								}

								if (!upstreamReg)
									upstreamReg = refReg;
								else if (refReg != upstreamReg || refMods != modifiers) {
									upstreamReg = 0;
									break;
								}

								swizzle += srcRegInfo.mComponents[refComp].mComp << 6;
								modifiers = srcRegInfo.mComponents[refComp].mMods;
							} else {
								swizzle += j << 6;
							}
						}

						if (!upstreamReg)
							break;

						src.mRegister = upstreamReg;
						src.mSwizzle = swizzle;
						src.mModifiers ^= modifiers;
					} while((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0);

					if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
						RegisterInfo& regInfo = regInfos[src.mRegister & kVDShaderRegIndexMask];
						for(int j=0; j<4; ++j) {
							if (srcmask & (1 << (i*4 + j))) {
								regInfo.mReferenceMask |= 1 << ((src.mSwizzle >> (2*j)) & 3);
							}
						}
					}
				}
			}

			if (insn.mDestRegister && (insn.mDestRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int rindex = insn.mDestRegister - kVDShaderRegR0;

				// remap temp register
				if (rindex >= mRegisterCount)
					return lexer.SetError("Internal compiler error: Invalid dest temp register index");

				int newrindex = tempRemap[rindex];
				if (newrindex < 0) {
					newrindex = newRegisterCount++;
					tempRemap[rindex] = newrindex;
				}

				rindex = newrindex;
				insn.mDestRegister = kVDShaderRegR0 + newrindex;

				RegisterInfo& dstRegInfo = regInfos[rindex];

				dstRegInfo.mWriteMask |= insn.mWriteMask;

				if (insn.mOp == kVDShaderOp_mov) {
					const VDShaderInsnSource& src1 = insn.mSources[0];

					for(int i=0; i<4; ++i) {
						if (insn.mWriteMask & (1 << i)) {
							dstRegInfo.mComponents[i].mReg = src1.mRegister;
							dstRegInfo.mComponents[i].mComp = (src1.mSwizzle >> (i+i)) & 3;
							dstRegInfo.mComponents[i].mMods = src1.mModifiers;
						}
					}
				} else {
					for(int i=0; i<4; ++i) {
						if (insn.mWriteMask & (1 << i))
							dstRegInfo.mComponents[i].mReg = 0;
					}
				}
			}
nop_insn:
			;
		}
		mConstants.swap(newConstants);

		// dead store elimination
		newinsns.clear();
		for(Insns::const_iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
			VDShaderInsn insn = *it;

			if (insn.mOp == kVDShaderOp_nop) {
				eliminatedAnything = true;
				continue;
			}

			if (insn.mDestRegister && (insn.mDestRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				RegisterInfo& dstRegInfo = regInfos[insn.mDestRegister - kVDShaderRegR0];

				if (insn.mWriteMask & ~dstRegInfo.mReferenceMask) {
					insn.mWriteMask &= dstRegInfo.mReferenceMask;
					eliminatedAnything = true;
				}
			}

			if (insn.mWriteMask)
				newinsns.push_back(insn);
			else
				eliminatedAnything = true;
		}

		mInsns.swap(newinsns);
	} while(eliminatedAnything);


	return true;
}

///////////////////////////////////////////////////////////////////////////////
VDShaderParser::VDShaderParser() {
}

VDShaderParser::~VDShaderParser() {
	while(!mFunctions.empty()) {
		delete mFunctions.back();
		mFunctions.pop_back();
	}
}

namespace {
	class NameRemapper {
	public:
		NameRemapper(VDShaderNameTable& dstNameTable, const VDShaderNameTable& srcNameTable) : mDstNameTable(dstNameTable), mSrcNameTable(srcNameTable) {}

		int Remap(int srcNameToken) {
			if (!srcNameToken)
				return 0;

			NameRemap::iterator it(mNameRemap.find(srcNameToken));

			if (it != mNameRemap.end())
				return it->second;

			const char *s = mSrcNameTable.GetName(srcNameToken);
			int dstNameToken = mDstNameTable.AddName(s, strlen(s));

			mNameRemap.insert(NameRemap::value_type(srcNameToken, dstNameToken));

			return dstNameToken;
		}

	protected:
		VDShaderNameTable& mDstNameTable;
		const VDShaderNameTable& mSrcNameTable;

		typedef std::map<int, int> NameRemap;
		NameRemap mNameRemap;
	};
}

bool VDShaderParser::Parse(const char *src, size_t len, const char *context, IVDShaderLogOutput *pOutput, VDShaderEffect **ppEffect) {
	mpEffect = new VDShaderEffect;

	mNextGlobalConstantRegister = 0;
	mCurrentScopeLevel = 0;
	mGlobalSamplers = 0;
	mpFunction = NULL;

	static const float kBaseConstants[]={
		0, 0, 0, 0
	};
	mConstants.assign(kBaseConstants, kBaseConstants + sizeof(kBaseConstants)/sizeof(kBaseConstants[0]));

	mLexer.Init(src, len, context, pOutput);

	while(TryParseGlobalStatement())
		;

	if (mLexer.IsErrorFlagSet())
		return NULL;

	NameRemapper nameRemapper(mpEffect->mNameTable, mNameTable);

	for(Functions::iterator it(mFunctions.begin()), itEnd(mFunctions.end()); it!=itEnd; ++it) {
		VDShaderParserFunction *pParserFunc = *it;

		VDShaderFunction *func = new VDShaderFunction;

		func->mConstants.swap(pParserFunc->mConstants);
		func->mInsns.swap(pParserFunc->mInsns);
		func->mParameterRegisterCount = pParserFunc->mParameterRegisterCount;
		func->mParameters.swap(pParserFunc->mParameters);
		func->mRegisterCount = pParserFunc->mRegisterCount;
		func->mReturnType = pParserFunc->mReturnType;

		for(VDShaderFunction::Parameters::iterator itP(func->mParameters.begin()), itPEnd(func->mParameters.end()); itP!=itPEnd; ++itP) {
			itP->mNameToken = nameRemapper.Remap(itP->mNameToken);
			itP->mSemanticToken = nameRemapper.Remap(itP->mSemanticToken);
		}

		func->mNameToken = nameRemapper.Remap(pParserFunc->mNameToken);

		mpEffect->mFunctions.push_back(func);

		delete pParserFunc;
	}

	mFunctions.clear();
	while(!mFunctions.empty()) {
		delete mFunctions.back();
		mFunctions.pop_back();
	}

	for(VDShaderEffect::Techniques::iterator it(mpEffect->mTechniques.begin()), itEnd(mpEffect->mTechniques.end()); it!=itEnd; ++it) {
		VDShaderTechnique& tech = **it;

		tech.mNameToken = nameRemapper.Remap(tech.mNameToken);

		for(VDShaderTechnique::Passes::iterator itP(tech.mPasses.begin()), itPEnd(tech.mPasses.end()); itP!=itPEnd; ++itP) {
			VDShaderPass& pass = **itP;

			pass.mNameToken = nameRemapper.Remap(pass.mNameToken);
		}
	}

	int numVars = mVarHeap.size();
	mpEffect->mVariables.resize(numVars);
	for(int i=0; i<numVars; ++i) {
		VDShaderVariable& var = mpEffect->mVariables[i];
		var = mVarHeap[i];
		var.mNameToken = nameRemapper.Remap(var.mNameToken);
		var.mSemanticToken = nameRemapper.Remap(var.mSemanticToken);
	}

	*ppEffect = mpEffect.release();
	return true;
}

bool VDShaderParser::GetErrorInfo(VDShaderErrorInfo& errorInfo) {
	return mLexer.GetErrorInfo(errorInfo);
}

bool VDShaderParser::TryParseGlobalStatement() {
	int tok = mLexer.Token();

	if (!tok)
		return false;

	if (tok == kTokenIdent) {
		int type = GetTypeFromIdent();

		if (type) {
			tok = mLexer.Token();
			int name = GetNameToken();

			tok = mLexer.Token();
			if (tok == '(') {
				if (!ParseFunction(type, name))
					return false;
				return true;
			} else {
				mLexer.Push();
				if (!ParseDeclarationList(type, name))
					return false;
				return true;
			}
		}

		if (mLexer.mTokenLength == 9 && !memcmp(mLexer.mpToken, "technique", 9))
			return ParseTechnique();
	}

	return mLexer.SetError("Expected declaration");
}

bool VDShaderParser::ParseTechnique() {
	vdautoptr<VDShaderTechnique> pTechnique(new VDShaderTechnique);

	int tok = mLexer.Token();

	pTechnique->mNameToken = 0;
	if (tok == kTokenIdent) {
		int name = GetNameToken();

		pTechnique->mNameToken = name;
		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected '{' after technique keyword");

	for(;;) {
		tok = mLexer.Token();

		if (tok == '}')
			break;

		if (tok != kTokenIdent || mLexer.mTokenLength != 4 || memcmp(mLexer.mpToken, "pass", 4))
			return mLexer.SetError("Expected pass or end of technique");

		if (!ParsePass(*pTechnique))
			return false;
	}

	mpEffect->mTechniques.push_back(pTechnique);
	pTechnique.release();

	return true;
}

bool VDShaderParser::ParsePass(VDShaderTechnique& tech) {
	vdautoptr<VDShaderPass> pPass(new VDShaderPass);

	int tok = mLexer.Token();

	pPass->mNameToken = 0;
	if (tok == kTokenIdent) {
		int name = GetNameToken();

		pPass->mNameToken = name;
		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected '{' after pass keyword");

	pPass->mPixelShaderIndex = -1;
	for(;;) {
		tok = mLexer.Token();

		if (tok == '}')
			break;

		if (tok == kTokenIdent) {
			if (mLexer.mTokenLength == 11 && !_memicmp(mLexer.mpToken, "PixelShader", 11)) {
				tok = mLexer.Token();

				if (tok != '=')
					return mLexer.SetError("Expected '=' after PixelShader keyword");

				if (!ParseExpression())
					return false;

				VDShaderParserVariable& v = mVarHeap.back();
				if (v.mType != kTypePixelShader)
					return mLexer.SetErrorF("Expression must be of type 'pixelshader', not '%s'", GetTypeName(v.mType));

				if (pPass->mPixelShaderIndex >= 0)
					return mLexer.SetError("Pixel shader already set for this pass");

				pPass->mPixelShaderIndex = v.mShader;

				mVarHeap.pop_back();

				tok = mLexer.Token();
				if (tok != ';')
					return mLexer.SetError("Expected ';' at end of pass statement");

				VDShaderParserFunction& func = *mFunctions[pPass->mPixelShaderIndex];
				if (!ValidateFunctionAsPixelShader(func))
					return false;

				continue;
			}
		}

		return mLexer.SetError("Expected pass statement or end of technique");
	}

	if (pPass->mPixelShaderIndex < 0)
		return mLexer.SetError("No pixel shader declared in pass");

	tech.mPasses.push_back(pPass);
	pPass.release();

	return true;
}

bool VDShaderParser::ParseFunction(int type, int name) {
	mpFunction = new VDShaderParserFunction;
	mpFunction->mNameToken = name;
	mpFunction->mReturnType = type;
	mpFunction->mIndex = mFunctions.size();
	mFunctions.push_back(mpFunction);

	int oldScopeLevel = mCurrentScopeLevel;
	mCurrentScopeLevel = (int)mVarHeap.size();

	// parse argument list
	int tok = mLexer.Token();
	if (tok != ')') {
		mLexer.Push();

		for(;;) {
			VDShaderParserVariable var;

			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Expected parameter type");

			if (mLexer.mTokenLength == 7 && !memcmp(mLexer.mpToken, "uniform", 7)) {
				var.mbUniform = true;
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Expected parameter type");
			}

			int type = GetTypeFromIdent();
			if (!type)
				return mLexer.SetError("Expected parameter type");

			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Expected parameter name");

			int paramName = GetNameToken();
			int paramSemantic = 0;

			tok = mLexer.Token();
			if (tok == ':') {
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Parameter semantic name must follow ':'");
				paramSemantic = GetNameToken();
				tok = mLexer.Token();
			}

			var.mNameToken = paramName;
			var.mSemanticToken = paramSemantic;
			var.mType = type;

			if (type == kTypeSampler) {
				var.mRegister = kVDShaderRegS0 + mpFunction->mSamplerCount++;
				mpFunction->mSamplerBindings.push_back(-1);
			} else
				var.mRegister = mpFunction->AllocateRegisters(kRowCountForType[type]);

			var.mbIsConst = false;

			mVarHeap.push_back(var);

			mpFunction->mParameters.push_back(var);

			if (tok == ')')
				break;
			else if (tok != ',')
				return mLexer.SetError("Expected ',' or ')'");
		}
	}

	mpFunction->mParameterRegisterCount = mpFunction->mRegisterCount;
	mpFunction->mReturnSemantic = 0;

	tok = mLexer.Token();
	if (tok == ':') {
		tok = mLexer.Token();
		if (tok != kTokenIdent)
			return mLexer.SetError("Expected return semantic type after ':'");
		mpFunction->mReturnSemantic = GetNameToken();
		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected function body");

	bool returnFound = false;
	if (!ParseBlock(returnFound))
		return false;

	if (!returnFound && mpFunction->mReturnType != kTypeVoid)
		return mLexer.SetError("Missing return statement");

	mVarHeap.resize(mCurrentScopeLevel);
	mCurrentScopeLevel = oldScopeLevel;

	if (!mpFunction->Optimize(mLexer, mNameTable))
		return false;

	return true;
}

bool VDShaderParser::ParseDeclarationList(int type) {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected type name");

	int name = GetNameToken();

	return ParseDeclarationList(type, name);
}

bool VDShaderParser::ParseDeclarationList(int type, int name) {
	int tok;

	if (type == kTypeVoid)
		return mLexer.SetError("Variables cannot be of type 'void'");

	for(;;) {
		int semantic = 0;
		tok = mLexer.Token();
		if (tok == ':') {
			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Variable semantic type must follow ':'");
			semantic = GetNameToken();
			tok = mLexer.Token();
		}

		int index = (int)mVarHeap.size();

		VDShaderParserVariable var;
		var.mNameToken = name;
		var.mSemanticToken = semantic;
		var.mType = type;
		var.mbIsConst = false;
		var.mRegister = 0;

		if (mpFunction) {
			if (type != kTypeSampler)
				var.mRegister = mpFunction->AllocateRegisters(kRowCountForType[type]);
			else
				return mLexer.SetError("Sampler variables may not be declared at function scope");
		} else {
			if (type == kTypeSampler)
				var.mRegister = kVDShaderRegGS0 + mGlobalSamplers++;
		}

		mVarHeap.push_back(var);

		if (tok == '=') {		
			if (!ParseExpression())
				return false;

			VDShaderParserVariable& var2 = mVarHeap.back();
			int varType = mVarHeap[index].mType;
			if (!ConvertToType(var2, varType))
				return mLexer.SetErrorF("Cannot convert initializer from type '%s' to '%s'", GetTypeName(var2.mType), GetTypeName(varType));

			var2.mNameToken = name;
			var2.mSemanticToken = semantic;
			mVarHeap[index] = var2;
			mVarHeap.pop_back();

			tok = mLexer.Token();
		}

		if (tok == ';')
			return true;

		if (tok != ',') {
			mLexer.SetError("Expected ',' or ';'");
			return false;
		}
	}
}

bool VDShaderParser::ParseBlock(bool& returnFound) {
	for(;;) {
		int tok = mLexer.Token();

		if (tok == '}')
			break;

		mLexer.Push();
		if (!ParseStatement(returnFound))
			return false;
	}

	return true;
}

bool VDShaderParser::ParseStatement(bool& returnFound) {
	int tok = mLexer.Token();

	if (tok == '{') {
		int oldScopeLevel = mCurrentScopeLevel;
		mCurrentScopeLevel = mVarHeap.size();

		bool result = ParseBlock(returnFound);

		mVarHeap.resize(oldScopeLevel);
		mCurrentScopeLevel = oldScopeLevel;
	} else if (tok == kTokenIdent) {
		int type = GetTypeFromIdent();

		if (type)
			return ParseDeclarationList(type);

		if (mLexer.mTokenLength == 6 && !memcmp(mLexer.mpToken, "return", 6)) {
			returnFound = true;
			return ParseReturnStatement();
		}

		int name = GetNameToken();

		int varLVIndex = FindLValue(name, false);
		if (varLVIndex < 0)
			return mLexer.SetErrorF("Undefined variable '%s'", GetName(name));

		bool writeMaskActive = false;
		uint8 dstWriteMask = 0;
		uint8 srcSwizzle = 0xE4;
		int dstType = mVarHeap[varLVIndex].mType;
		int dstCols = kColCountForType[dstType];

		tok = mLexer.Token();
		if (tok == '.') {
			// parse write mask
			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Expected destination write mask for assignment after '.'");

			if (kRowCountForType[dstType] != 1)
				return mLexer.SetErrorF("Cannot use destination write mask with type '%s'", GetTypeName(dstType));

			int len = mLexer.mTokenLength;
			if (len > dstCols)
				return mLexer.SetError("Destination write mask too long");

			for(int i=0; i<len; ++i) {
				char c = mLexer.mpToken[i];
				int component;

				switch(c) {
					case 'x':
					case 'r':
						component = 0;
						break;
					case 'y':
					case 'g':
						component = 1;
						break;
					case 'z':
					case 'b':
						component = 2;
						break;
					case 'w':
					case 'a':
						component = 3;
						break;
					default:
						return mLexer.SetErrorF("Invalid destination write mask '%.*s'", len, mLexer.mpToken);
				}

				uint8 bit = (1 << component);
				if (dstWriteMask & bit)
					return mLexer.SetErrorF("Cannot repeat component '%c' in write mask", c);
				dstWriteMask |= bit;

				int comp2x = component+component;
				srcSwizzle = (srcSwizzle & ~(3 << comp2x)) + (i << comp2x);
			}

			writeMaskActive = true;

			tok = mLexer.Token();
		}

		if (tok != '=')
			return mLexer.SetError("Expected '=' for assignment statement");

		if (!ParseExpression())
			return false;

		VDShaderParserVariable& lvar = mVarHeap[varLVIndex];
		VDShaderParserVariable& rvar = mVarHeap.back();
		if (!ConvertToType(rvar, lvar.mType))
			return mLexer.SetErrorF("Cannot convert from type '%s' to type '%s'", GetTypeName(rvar.mType), GetTypeName(lvar.mType));

		if (writeMaskActive) {
			int newReg = mpFunction->AllocateRegisters(1);

			mpFunction->PromoteToConstantRegisters(lvar);
			mpFunction->PromoteToConstantRegisters(rvar);

			int validMask = (1 << dstCols) - 1;
			int preserveMask = validMask ^ dstWriteMask;

			if (preserveMask)
				mpFunction->mov(newReg, preserveMask, lvar.mRegister);

			mpFunction->mov_swizzle(newReg, dstWriteMask, rvar.mRegister, srcSwizzle);

			lvar.mRegister = newReg;
		} else {
			int sema = lvar.mSemanticToken;
			lvar = rvar;
			lvar.mNameToken = name;
			lvar.mSemanticToken = sema;
		}

		tok = mLexer.Token();
		if (tok != ';')
			return mLexer.SetError("Expected end of statement");

		return true;
	}

	return mLexer.SetError("Expected statement");
}

bool VDShaderParser::ParseReturnStatement() {
	if (!ParseExpression())
		return false;

	VDShaderParserVariable& var = mVarHeap.back();

	if (var.mbIsConst)
		mpFunction->PromoteToConstantRegisters(var);

	if (!ConvertToType(var, mpFunction->mReturnType))
		return mLexer.SetErrorF("Cannot convert return value from type '%s' to type '%s'", GetTypeName(var.mType), GetTypeName(mpFunction->mReturnType));

	int rows = kRowCountForType[mpFunction->mReturnType];
	for(int i=0; i<rows; ++i)
		mpFunction->mov(kVDShaderRegO0 + i, 0x0F, var.mRegister + i);
	mVarHeap.pop_back();

	int tok = mLexer.Token();
	if (tok != ';')
		return mLexer.SetError("Expected end of return statement");

	return true;
}

namespace {
	enum ExpressionOp {
		kOpNone,
		kOpAdd,
		kOpSub,
		kOpMul,
		kOpDiv,
		kOpNegate,
		kOpSwizzle,
	};
}

bool VDShaderParser::ParseExpression() {
	size_t opBase = mOpStack.size();
	int stackBase = mVarHeap.size();
	bool expectValue = true;
	int parenLevel = 0;

	for(;;) {
		int tok = mLexer.Token();

		if (expectValue) {
			if (tok == '+') {		// unary +
				const VDShaderParserVariable& var = mVarHeap.back();

				if (!kRowCountForType[var.mType])
					return mLexer.SetError("Unary '+' can only be applied to numeric types");
				continue;
			} else if (tok == '-') {
				const VDShaderParserVariable& var = mVarHeap.back();

				if (!kRowCountForType[var.mType])
					return mLexer.SetError("Unary '-' can only be applied to numeric types");

				mOpStack.push_back(kOpNegate + (9<<8));
				continue;
			} else if (tok == '(') {
				parenLevel += 0x10000;
			} else if (tok == kTokenIdent) {
				int type = GetTypeFromIdent();
				if (type) {
					if (!ParseCtorExpression(type))
						return false;

					expectValue = false;
					continue;
				}

				if (mLexer.mTokenLength == 13 && !memcmp(mLexer.mpToken, "sampler_state", 13)) {
					if (!ParseSamplerState())
						return false;

					expectValue = false;
					continue;
				}

				if (mLexer.mTokenLength == 7 && !memcmp(mLexer.mpToken, "compile", 7)) {
					if (!ParseCompileExpression())
						return false;

					expectValue = false;
					continue;
				}

				int varName = GetNameToken();
				const VDShaderIntrinsicEntry *pBaseIntrinsic = VDShaderParserGetIntrinsicByName(GetName(varName));

				tok = mLexer.Token();
				if (tok == '(') {
					int argCount = 0;
					int argBase = mVarHeap.size();

					tok = mLexer.Token();
					if (tok != ')') {
						mLexer.Push();

						for(;;) {
							++argCount;
							if (!ParseExpression())
								return false;

							tok = mLexer.Token();
							if (tok == ')')
								break;

							if (tok != ',')
								return mLexer.SetError("Syntax error in argument list");
						}
					}

					if (pBaseIntrinsic) {
						if (!argCount) {
							VDShaderParserVariable var;
							var.mNameToken = 0;
							var.mSemanticToken = 0;
							var.mRegister = 0;
							var.mType = 0;
							var.mbIsConst = false;
							mVarHeap.push_back(var);
						}

						if (!VDShaderParserCompileIntrinsic(mLexer, *mpFunction, pBaseIntrinsic, &mVarHeap[argBase], argCount))
							return false;

						if (argCount > 1)
							mVarHeap.resize(argBase + 1);

						expectValue = false;
						continue;
					} else {
						Functions::const_iterator it(mFunctions.begin()), itEnd(mFunctions.end());
						for(; it!=itEnd; ++it) {
							VDShaderParserFunction& calledFunc = **it;
							if (calledFunc.mNameToken == varName) {
								// check argument count
								int expectedArgCount = calledFunc.mParameters.size();
								if (argCount != expectedArgCount)
									return mLexer.SetErrorF("Expected %d arguments to function '%s', found %d\n", expectedArgCount, GetName(varName), argCount);

								// coerce arguments
								for(int i=0; i<argCount; ++i) {
									VDShaderParserVariable& arg = mVarHeap[argBase + i];
									const VDShaderVariable& argVar = calledFunc.mParameters[i];

									if (!ConvertToType(arg, argVar.mType))
										return mLexer.SetErrorF("Cannot convert argument %d from type '%s' to type '%s'", i+1, GetTypeName(arg.mType), GetTypeName(argVar.mType));
								}

								// inline the function
								VDShaderParserVariable returnValue;
								if (!mpFunction->Inline(mLexer, calledFunc, argCount ? &mVarHeap[argBase] : NULL, returnValue))
									return false;

								returnValue.mbIsConst = false;

								mVarHeap.resize(argBase + 1);
								mVarHeap.back() = returnValue;

								expectValue = false;
								break;
							}
						}

						if (it == itEnd)
							return mLexer.SetErrorF("Undefined function '%s'", GetName(varName));
						continue;
					}
				} else {
					mLexer.Push();

					if (pBaseIntrinsic)
						return mLexer.SetErrorF("Expected argument list for intrinsic '%s'", pBaseIntrinsic->mpName);
				}

				int varIndex = FindVariable(varName, false);
				if (varIndex < 0)
					return mLexer.SetErrorF("Undefined variable '%s'", GetName(varName));

				const VDShaderParserVariable& varRef = mVarHeap[varIndex];

				VDShaderParserVariable varCopy(varRef);

				if (varCopy.mType == kTypeSampler) {
					if ((varCopy.mRegister & kVDShaderRegTypeMask) == kVDShaderRegGS0) {
						if (!mpFunction) {
							return mLexer.SetError("Cannot refer to global samplers at global scope");
						} else {
							int gsindex = varCopy.mRegister & kVDShaderRegIndexMask;

							varCopy.mRegister = mpFunction->AllocateSampler(gsindex);
						}
					}
				} else {
					if (!varCopy.mbIsConst && !varCopy.mRegister)
						return mLexer.SetErrorF("Reference to uninitialized variable '%s'", GetName(varName));
				}

				mVarHeap.push_back(varCopy);
				expectValue = false;
			} else if (tok == kTokenFloat || tok == kTokenInt) {
				VDShaderParserVariable constVar;
				constVar.mNameToken = 0;
				constVar.mSemanticToken = 0;
				constVar.mRegister = 0;
				constVar.mType = kTypeFloat;
				constVar.mbIsConst = true;
				memset(constVar.m, 0, sizeof constVar.m);

				if (tok == kTokenFloat)
					constVar.m[0].v[0] = mLexer.mFloatVal;
				else
					constVar.m[0].v[0] = (float)mLexer.mIntVal;

				mVarHeap.push_back(constVar);
				expectValue = false;
			} else {
				return mLexer.SetError("Expected expression value");
			}
		} else {
			int op = 0;
			if (tok == '+')
				op = kOpAdd + (1 << 8) + parenLevel;
			else if (tok == '-')
				op = kOpSub + (1 << 8) + parenLevel;
			else if (tok == '*')
				op = kOpMul + (2 << 8) + parenLevel;
			else if (tok == '/')
				op = kOpDiv + (2 << 8) + parenLevel;
			else if (tok == '.')
				op = kOpSwizzle + (10 << 8) + parenLevel;
			else if (tok == ')') {
				if (parenLevel) {
					parenLevel -= 0x10000;
					continue;
				}
			}

			// reduce as necessary
			while(mOpStack.size() > opBase) {
				int otherop = mOpStack.back();

				if (op > otherop)
					break;

				mOpStack.pop_back();

				ExpressionOp baseop = (ExpressionOp)(otherop & 0xff);

				if (baseop == kOpNegate) {
					VDShaderParserVariable& v = mVarHeap.back();

					if (v.mbIsConst) {
						v.m[0] = -v.m[0];
						v.m[1] = -v.m[1];
						v.m[2] = -v.m[2];
						v.m[3] = -v.m[3];
					} else {
						int rows = kRowCountForType[v.mType];
						unsigned char writeMask = (unsigned char)((1 << kColCountForType[v.mType]) - 1);
						int r = mpFunction->AllocateRegisters(rows);

						for(int i=0; i<rows; ++i)
							mpFunction->mov_mods(r+i, writeMask, v.mRegister, 0xE4, kVDShaderMod_Negate);

						v.mRegister = r;
					}
				} else {
					// coerce
					VDShaderParserVariable& var1 = *(mVarHeap.end() - 2);
					VDShaderParserVariable& var2 = mVarHeap.back();

					int commonType = GetBinaryResultType(var1.mType, var2.mType);
					if (!commonType)
						return mLexer.SetErrorF("Cannot operate between types '%s' and '%s'", GetTypeName(var1.mType), GetTypeName(var2.mType));

					ConvertToType(var1, commonType);
					ConvertToType(var2, commonType);

					if (var1.mbIsConst != var2.mbIsConst) {
						mpFunction->PromoteToConstantRegisters(var1);
						mpFunction->PromoteToConstantRegisters(var2);
					}

					unsigned char writeMask = (unsigned char)((1 << kColCountForType[commonType]) - 1);

					int rowCount = kRowCountForType[commonType];
					int colCount = kRowCountForType[commonType];
					int srcBase1 = var1.mRegister;
					int srcBase2 = var2.mRegister;
					int dstBase = 0;
					
					if (!var1.mbIsConst)
						dstBase = mpFunction->AllocateRegisters(rowCount);

					switch(baseop) {
						case kOpAdd:
							if (!var1.mbIsConst) {
								for(int row=0; row<rowCount; ++row)
									mpFunction->add(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
							} else {
								for(int row=0; row<rowCount; ++row)
									var1.m[row] += var2.m[row];
							}
							break;
						case kOpSub:
							if (!var1.mbIsConst) {
								for(int row=0; row<rowCount; ++row)
									mpFunction->sub(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
							} else {
								for(int row=0; row<rowCount; ++row)
									var1.m[row] -= var2.m[row];
							}
							break;
						case kOpMul:
							if (!var1.mbIsConst) {
								for(int row=0; row<rowCount; ++row)
									mpFunction->mul(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
							} else {
								for(int row=0; row<rowCount; ++row)
									var1.m[row] *= var2.m[row];
							}
							break;
						case kOpDiv:
							if (!var1.mbIsConst) {
								int tmp = mpFunction->AllocateRegisters(1);

								for(int row=0; row<rowCount; ++row) {
									for(unsigned char bit=1; bit<16; bit+=bit) {
										if (writeMask & bit)
											mpFunction->rcp(tmp, bit, srcBase2+row, bit);
									}

									mpFunction->mul(dstBase+row, writeMask, srcBase1+row, tmp);
								}
							} else {
								switch(colCount) {
								case 1:
									for(int row=0; row<rowCount; ++row)
										var1.m[row] = div1(var1.m[row], var2.m[row]);
									break;
								case 2:
									for(int row=0; row<rowCount; ++row)
										var1.m[row] = div2(var1.m[row], var2.m[row]);
									break;
								case 3:
									for(int row=0; row<rowCount; ++row)
										var1.m[row] = div3(var1.m[row], var2.m[row]);
									break;
								case 4:
									for(int row=0; row<rowCount; ++row)
										var1.m[row] = div4(var1.m[row], var2.m[row]);
									break;
								}
							}
							break;
					}

					var1.mType = commonType;
					var1.mRegister = dstBase;
					var1.mNameToken = 0;
					var1.mSemanticToken = 0;

					mVarHeap.pop_back();
				}
			}

			expectValue = true;

			if (!op)
				break;

			if ((op & 0xff) != kOpSwizzle) {
				mOpStack.push_back(op);
			} else {
				VDShaderParserVariable& v = mVarHeap.back();
				int maxComponents;
				switch(v.mType) {
					case kTypeFloat:
						maxComponents = 1;
						break;
					case kTypeFloat2:
						maxComponents = 2;
						break;
					case kTypeFloat3:
						maxComponents = 3;
						break;
					case kTypeFloat4:
						maxComponents = 4;
						break;
					default:
						return mLexer.SetErrorF("Cannot swizzle type '%s'", GetTypeName(v.mType));
				}

				// parse swizzle
				unsigned char swizzle;
				int swizzleLength;

				if (!ParseSwizzle(v.mType, maxComponents, swizzle, swizzleLength))
					return false;

				unsigned char mask = 0xFF >> (8 - 2*std::min(maxComponents, swizzleLength));

				if (((swizzle ^ 0xE4) & mask) || swizzleLength > maxComponents) {
					if (v.mbIsConst) {
						v.m[0] = nsVDShaderArith::swizzle(v.m[0], swizzle);
					} else {
						int rA = mpFunction->AllocateRegisters(1);
						VDShaderConstant c = {0};
						int rZ = mpFunction->AddConstant(c);
						switch(swizzleLength) {
							case 1:
								mpFunction->mov_swizzle(rA, 0x01, v.mRegister, swizzle);
								mpFunction->mov(rA, 0x0E, rZ);
								break;
							case 2:
								mpFunction->mov_swizzle(rA, 0x03, v.mRegister, swizzle);
								mpFunction->mov(rA, 0x0C, rZ);
								break;
							case 3:
								mpFunction->mov_swizzle(rA, 0x07, v.mRegister, swizzle);
								mpFunction->mov(rA, 0x08, rZ);
								break;
							case 4:
								mpFunction->mov_swizzle(rA, 0x0F, v.mRegister, swizzle);
								break;
						}

						v.mRegister = rA;
					}
				}

				v.mType = kTypeFloat + swizzleLength - 1;
				v.mNameToken = 0;
				v.mSemanticToken = 0;
				expectValue = false;
			}
		}
	}

	if (parenLevel)
		return mLexer.SetError("Unmatched '(' in expression");

	mLexer.Push();
	return true;
}

bool VDShaderParser::ParseSwizzle(int type, int maxComponents, unsigned char& swizzle, int& swizzleLength) {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected swizzle");

	swizzle = 0;
	swizzleLength = mLexer.mTokenLength;

	if (swizzleLength > 4)
		return mLexer.SetError("Invalid swizzle");

	for(int i=0; i<swizzleLength; ++i) {
		swizzle >>= 2;

		int component;
		switch(mLexer.mpToken[i]) {
			case 'x':	case 'r':	component = 0; break;
			case 'y':	case 'g':	component = 1; break;
			case 'z':	case 'b':	component = 2; break;
			case 'w':	case 'a':	component = 3; break;
			default:
				return mLexer.SetError("Invalid swizzle");
		}

		if (component > maxComponents)
			return mLexer.SetErrorF("Invalid swizzle component for type '%s'", GetTypeName(type));

		swizzle += component << 6;
	}

	for(int i=swizzleLength; i<4; ++i)
		swizzle = (swizzle & 0xc0) + (swizzle >> 2);

	return true;
}

bool VDShaderParser::ParseSamplerState() {
	int tok = mLexer.Token();

	if (tok != '{')
		return mLexer.SetError("'sampler_state' must be followed by '{'");

	for(;;) {
		tok = mLexer.Token();
		if (tok == '}')
			break;

		if (tok != kTokenIdent)
			return mLexer.SetError("Expected sampler state");

		// parse sampler state

		tok = mLexer.Token();
		if (tok != '=')
			return mLexer.SetError("Expected '=' after sampler state name");

		// parse sampler value
		int basetok = mLexer.Token();
		if (basetok == '(' || basetok == '<') {
			ParseExpression();
			mVarHeap.pop_back();

			int matchtok = (basetok == '(') ? ')' : '>';
			tok = mLexer.Token();
			if (tok != matchtok)
				return mLexer.SetErrorF("Expected '%c' after sampler state expression", matchtok);
		} else if (basetok == kTokenFloat) {
		} else if (basetok == kTokenInt) {
		} else if (basetok == kTokenIdent) {
		}

		tok = mLexer.Token();
		if (tok != ';')
			return mLexer.SetError("Expected ';' after sampler state declaration");
	}

	return true;
}

bool VDShaderParser::ParseCompileExpression() {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected shader profile after 'compile' keyword");

	tok = mLexer.Token();
	if (tok != kTokenIdent)
		return mLexer.SetError("Function name expected");

	int nameToken = GetNameToken();

	VDShaderParserFunction *pFunc = NULL;
	for(Functions::const_iterator it(mFunctions.begin()), itEnd(mFunctions.end()); it!=itEnd; ++it) {
		VDShaderParserFunction& calledFunc = **it;
		if (calledFunc.mNameToken == nameToken) {
			pFunc = &calledFunc;
			break;
		}
	}

	if (!pFunc)
		return mLexer.SetErrorF("Unknown function '%s'", GetName(nameToken));

	// start new function
	mpFunction = new VDShaderParserFunction;
	mpFunction->mNameToken = pFunc->mNameToken;
	mpFunction->mReturnType = pFunc->mReturnType;
	mpFunction->mReturnSemantic = pFunc->mReturnSemantic;
	mpFunction->mIndex = mFunctions.size();
	mFunctions.push_back(mpFunction);

	VDShaderFunction::Parameters::const_iterator itPBegin(pFunc->mParameters.begin()), itPEnd(pFunc->mParameters.end()), itP(itPBegin);

	// copy over and allocate all non-uniform parameters
	int uniformArgCount = 0;
	for(itP = itPBegin; itP != itPEnd; ++itP) {
		const VDShaderVariable& parm = *itP;
		if (parm.mbUniform) {
			++uniformArgCount;
			continue;
		}

		VDShaderVariable newParm(parm);

		if (int rows = kRowCountForType[parm.mType])
			newParm.mRegister = mpFunction->AllocateRegisters(rows);

		mpFunction->mParameters.push_back(newParm);
	}

	mpFunction->mParameterRegisterCount = mpFunction->mRegisterCount;

	tok = mLexer.Token();
	if (tok != '(')
		return mLexer.SetError("Expected '(' after function name");

	int varBase = mVarHeap.size();
	int parmIndex = 0;

	tok = mLexer.Token();
	itP = itPBegin;
	bool endOfArgs = true;

	if (tok != ')') {
		mLexer.Push();
		endOfArgs = false;
	}

	while(itP != itPEnd) {
		const VDShaderVariable& parmvar = *itP++;
		int rows = kRowCountForType[parmvar.mType];
		int cols = kColCountForType[parmvar.mType];
		unsigned char mask = (1 << cols) - 1;

		// if we have a non-uniform parameter, alias it against our argument
		if (!parmvar.mbUniform) {
			VDShaderParserVariable nonUniParm;
			nonUniParm.mNameToken = parmvar.mNameToken;
			nonUniParm.mSemanticToken = parmvar.mSemanticToken;
			nonUniParm.mType = parmvar.mType;
			nonUniParm.mRegister = parmvar.mRegister;
			nonUniParm.mbIsConst = false;
			mVarHeap.push_back(nonUniParm);
			continue;
		}

		if (endOfArgs)
			return mLexer.SetErrorF("Function has %d uniform arguments; only %d provided", uniformArgCount, parmIndex);

		// uniform argument - add to constants
		if (!ParseExpression())
			return false;

		VDShaderParserVariable& arg = mVarHeap.back();

		if (!arg.mbIsConst && arg.mType != kTypeSampler)
			return mLexer.SetError("Function argument must be a constant expression");

		if (!ConvertToType(arg, parmvar.mType))
			return mLexer.SetErrorF("Cannot convert parameter %d from type '%s' to type '%s'", parmIndex+1, GetTypeName(arg.mType), GetTypeName(parmvar.mType));

		VDShaderParserVariable argVar;
		argVar.mbIsConst = false;
		argVar.mbUniform = false;
		argVar.mNameToken = 0;
		argVar.mSemanticToken = 0;
		argVar.mType = arg.mType;
		for(int row=0; row<rows; ++row) {
			int r = mpFunction->AddConstant(arg.m[row]);

			if (!row)
				argVar.mRegister = r;
		}

		++parmIndex;

		tok = mLexer.Token();
		if (tok == ')')
			endOfArgs = true;
		else if (tok != ',')
			return mLexer.SetError("Expected ',' or ')' in function argument list");
	}

	if (!endOfArgs)
		return mLexer.SetErrorF("Too many arguments for function '%s'", GetName(pFunc->mNameToken));

	// specialize the function
	VDShaderVariable retVar;
	if (!mpFunction->Inline(mLexer, *pFunc, itPBegin != itPEnd ? &mVarHeap[varBase] : NULL, retVar))
		return false;

	int retrows = kRowCountForType[retVar.mType];
	int retcols = kColCountForType[retVar.mType];
	for(int row=0; row<retrows; ++row)
		mpFunction->mov(kVDShaderRegO0 + row, (1 << retcols) - 1, retVar.mRegister + row);

	if (!mpFunction->Optimize(mLexer, mNameTable))
		return false;

	mVarHeap.resize(varBase);

	// construct pixel shader variable
	VDShaderParserVariable v;
	v.mNameToken = 0;
	v.mSemanticToken = 0;
	v.mRegister = 0;
	v.mType = kTypePixelShader;
	v.mbIsConst = true;
	v.mShader = mpFunction->mIndex;
	mVarHeap.push_back(v);

	mpFunction = NULL;
	return true;
}

bool VDShaderParser::ParseCtorExpression(int type) {
	if (type == kTypeVoid)
		return mLexer.SetError("Cannot create value of type 'void'");

	int tok = mLexer.Token();
	if (tok != '(')
		return mLexer.SetError("Expected '(' after type");

	int rows = kRowCountForType[type];
	int cols = kColCountForType[type];

	VDShaderParserVariable newVar;
	newVar.mNameToken = 0;
	newVar.mSemanticToken = 0;
	newVar.mRegister = 0;
	newVar.mType = type;
	memset(newVar.m, 0, sizeof newVar.m);

	uint32 constMask = 0;
	bool nonConst = false;

	for(int row=0; row<rows; ++row) {
		int col = 0;

		while(col < cols) {
			if (row+col) {
				tok = mLexer.Token();
				if (tok != ',')
					return mLexer.SetError("Expected ',' in vector constructor argument list");
			}

			if (!ParseExpression())
				return false;

			const VDShaderParserVariable argVar(mVarHeap.back());
			mVarHeap.pop_back();

			int argRows = kRowCountForType[argVar.mType];
			int argCols = kColCountForType[argVar.mType];

			if (argRows > 1)
				return mLexer.SetError("Cannot initialize with a multi-row matrix");

			if (argCols + col > 4)
				return mLexer.SetErrorF("Too many elements in row %d", row);

			if (argVar.mbIsConst) {
				memcpy(&newVar.m[row].v[col], argVar.m[0].v, sizeof(float)*argCols);
				constMask |= ((1 << argCols) - 1) << (col + row*4);
			} else {
				if (!nonConst)
					newVar.mRegister = mpFunction->AllocateRegisters(rows);

				mpFunction->mov_swizzle(newVar.mRegister + row, ((1 << argCols) - 1) << col, argVar.mRegister, 0xE4 << (col+col));
				nonConst = true;
			}

			col += argCols;
		}
	}

	if (nonConst) {
		newVar.mbIsConst = false;

		for(int row=0; row<rows; ++row) {
			unsigned char mask = (constMask >> (4*row)) & 15;

			if (mask) {
				int r = mpFunction->AddConstant(newVar.m[row]);

				mpFunction->mov(newVar.mRegister + row, mask, r);
			}
		}
	} else {
		newVar.mbIsConst = true;
	}

	mVarHeap.push_back(newVar);

	tok = mLexer.Token();
	if (tok == ',')
		return mLexer.SetError("Too many arguments to vector constructor");

	if (tok != ')')
		return mLexer.SetError("Expected ')' after ctor-expression");

	return true;
}

int VDShaderParser::GetTypeFromIdent() const {
	const char *s = mLexer.mpToken;

	switch(mLexer.mTokenLength) {
		case 4:
			if (!memcmp(s, "void", 4))
				return kTypeVoid;
			break;
		case 5:
			if (!memcmp(s, "float", 5))
				return kTypeFloat;
			break;
		case 6:
			if (!memcmp(s, "float", 5)) {
				if (s[5] == '2')
					return kTypeFloat2;
				else if (s[5] == '3')
					return kTypeFloat3;
				else if (s[5] == '4')
					return kTypeFloat4;
			}
			break;
		case 7:
			if (!memcmp(s, "sampler", 7))
				return kTypeSampler;
			break;
		case 8:
			if (!memcmp(s, "float", 5) && s[6] == 'x') {
				unsigned rows = (unsigned)(s[5] - '0');
				unsigned cols = (unsigned)(s[7] - '0');

				if ((rows - 2) < 3 && (cols - 2) < 3)
					return kTypeFloat2x2 + (rows*3) + cols - 8;
			}
			break;
		case 11:
			if (!memcmp(s, "pixelshader", 11))
				return kTypePixelShader;
			break;
		case 12:
			if (!memcmp(s, "vertexshader", 12))
				return kTypeVertexShader;
			break;
	}

	return kTypeInvalid;
}

int VDShaderParser::GetNameToken() {
	int token = mNameTable.AddName(mLexer.mpToken, mLexer.mTokenLength);

	return token;
}

const char *VDShaderParser::GetTypeName(int type) const {
	return VDShaderGetTypeName(type);
}

const char *VDShaderParser::GetName(int name) const {
	return mNameTable.GetName(name);
}

int VDShaderParser::FindVariable(int name, bool localOnly) {
	int lvindex = FindLValue(name, localOnly);

	return lvindex;
}

int VDShaderParser::FindLValue(int name, bool localOnly) {
	int startOffset = 0;

	if (localOnly)
		startOffset = mCurrentScopeLevel;

	for(VarHeap::const_iterator it(mVarHeap.begin() + startOffset), itEnd(mVarHeap.end()); it!=itEnd; ++it) {
		const VDShaderVariable& var = *it;

		if (var.mNameToken == name)
			return it - mVarHeap.begin();
	}

	return -1;
}

bool VDShaderParser::ConvertToType(VDShaderParserVariable& var, int dstType) {
	int srcType = var.mType;

	if (srcType == dstType)
		return true;

	int srcRows = kRowCountForType[srcType];
	int srcCols = kColCountForType[srcType];
	unsigned char srcswizzle = 0xE4;
	if (srcCols == 1) {
		srcswizzle = 0;
		srcCols = 4;
	}

	int dstRows = kRowCountForType[dstType];
	int dstCols = kColCountForType[dstType];

	if (!srcRows || !dstRows)
		return false;

	int minRows = srcRows < dstRows ? srcRows : dstRows;
	int minCols = srcCols < dstCols ? srcCols : dstCols;

	if (!var.mbIsConst) {
		int srcBase = var.mRegister;
		int dstBase = mpFunction->AllocateRegisters(dstRows);

		unsigned char copymask = (1 << minCols) - 1;
		unsigned char zeromask = ((1 << dstCols) - 1) & ~copymask;

		int czero = mpFunction->AddConstantZero();

		for(int row=0; row<minRows; ++row) {
			if (copymask)
				mpFunction->mov_swizzle(dstBase+row, copymask, srcBase+row, srcswizzle);
			if (zeromask)
				mpFunction->mov(dstBase+row, zeromask, czero);
		}

		for(int row=minRows; row<dstRows; ++row)
			mpFunction->mov(dstBase+row, 0xF, czero);

		var.mRegister = dstBase;
	} else {
		if (srcswizzle == 0) {
			for(int row=0; row<minRows; ++row) {
				VDShaderConstant& c = var.m[row];
				for(int col=1; col<dstCols; ++col)
					c.v[col] = c.v[0];
			}
		} else if (dstCols < srcCols) {
			for(int row=0; row<minRows; ++row)
				memset(&var.m[row].v[dstCols], 0, sizeof(float)*(srcCols - dstCols));
		}

		if (dstRows < srcRows)
			memset(&var.m[dstRows].v, 0, sizeof(float)*4*(srcRows - dstRows));
	}

	var.mType = dstType;
	return true;
}

bool VDShaderParser::ValidateFunctionAsPixelShader(const VDShaderParserFunction& func) {
	// check return types
	if (func.mReturnType != kTypeFloat4)
		return mLexer.SetErrorF("'%s': Pixel shader must return float4", GetName(func.mNameToken));

	if (strcmp(GetName(func.mReturnSemantic), "COLOR0"))
		return mLexer.SetErrorF("'%s': Pixel shader return value must have COLOR0 semantic", GetName(func.mNameToken));

	// check arguments
	int texcoord0Token = mNameTable.GetName("TEXCOORD0", 9);

	for(VDShaderParserFunction::Parameters::const_iterator it(func.mParameters.begin()), itEnd(func.mParameters.end()); it!=itEnd; ++it) {
		const VDShaderVariable& var = *it;

		if (!var.mSemanticToken)
			return mLexer.SetErrorF("Invalid pixel shader function '%s': missing semantic on parameter '%s'", GetName(func.mNameToken), GetName(var.mNameToken));

		if (var.mSemanticToken != texcoord0Token)
			return mLexer.SetErrorF("Unsupported semantic for pixel shader; must be 'TEXCOORD0'");
	}

	return true;
}
