//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "x86disasm.h"

#ifdef _M_AMD64
	#define VD_PTR_08lx	"%08I64x"
#else
	#define VD_PTR_08lx	"%08x"
#endif

namespace {
	char *strtack(char *s, const char *t) {
		while(*s = *t)
			++s, ++t;

		return s;
	}

	// 1-15 are static lookups
	static const char kTarget_r32		= 1;
	static const char kTarget_r16		= 2;
	static const char kTarget_r8		= 3;
	static const char kTarget_rm		= 4;
	static const char kTarget_rx		= 5;
	static const char kTarget_rc		= 6;
	static const char kTarget_rd		= 7;
	static const char kTarget_rs		= 8;
	static const char kTarget_rf		= 9;

	// 16-31 are dynamic translations
	static const char kTarget_r1632		= 16;
	static const char kTarget_rmx		= 17;
	static const char kTarget_x			= 18;
	static const char kTarget_hx		= 19;
	static const char kTarget_lx		= 20;
	static const char kTarget_s			= 21;
	static const char kTarget_o			= 22;
	static const char kTarget_ho		= 23;
	static const char kTarget_lo		= 24;
	static const char kTarget_a			= 25;
	static const char kTarget_ha		= 26;
	static const char kTarget_la		= 27;
	static const char kTarget_r3264		= 28;
	static const char kTarget_r163264	= 29;
	static const char kTarget_ext		= 30;

	static const char kTarget_ext_r3264rexX	= 1;
	static const char kTarget_ext_r3264rexB = 2;

	static const char kTarget_ap		= (char)224;
	static const char kTarget_p_cs		= (char)225;
	static const char kTarget_p_ss		= (char)226;
	static const char kTarget_p_ds		= (char)227;
	static const char kTarget_p_es		= (char)228;
	static const char kTarget_p_fs		= (char)229;
	static const char kTarget_p_gs		= (char)230;
	static const char kTarget_p_66		= (char)231;
	static const char kTarget_p_67		= (char)232;
	static const char kTarget_p_F2		= (char)233;
	static const char kTarget_p_F3		= (char)234;
	static const char kTarget_p_rex		= (char)235;

	long VDDisasmPack32(const ptrdiff_t *src) {
		return src[0] + (src[1]<<8) + (src[2]<<16) + (src[3]<<24);
	}
}

///////////////////////////////////////////////////////////////////////////

VDDisassemblyContext::VDDisassemblyContext()
	: mpAllocBlock(NULL)
	, mpSymLookup(NULL)
{
}

VDDisassemblyContext::~VDDisassemblyContext() {
	free(mpAllocBlock);
}

bool VDDisassemblyContext::Init(const void *data) {
	// verify header
	const char *src = (const char *)data;

	if (src[0] != '[' || src[3] != '|')
		return false;

#ifdef _M_AMD64
	if (memcmp((char *)src + 6, "] VirtualDub disasm module (AMD64:", 34))
		return false;
#else
	if (memcmp((char *)src + 6, "] VirtualDub disasm module (IA32:", 33))
		return false;
#endif

	// Check version number
	int compat_version = (src[4]-'0')*10 + (src[5] - '0');

	if (compat_version > 2)
		return false;	// resource is too new for us to load

	long packSize = *(long *)((char *)data + 64);
	long depackSize = *(long *)((char *)data + 68);

	mpAllocBlock = malloc(depackSize + 65536);

	pRuleBase = (const unsigned char *)mpAllocBlock;
	pRuleSystemOffsets = (const unsigned *)pRuleBase;
	stack = (ptrdiff_t *)((char *)pRuleBase + depackSize);
	mpHeap = (char *)stack + 128;

	Decompress((void *)pRuleBase, (unsigned char *)data + 72, packSize);

	return true;
}

const char *VDDisassemblyContext::Disassemble(const unsigned char *source, int bytes, int& count) {
	const unsigned char *src2 = source;
	const unsigned char *src_end;
	char *s;

	mbAddressOverride = false;
	mbSizeOverride = false;
	mbRepePrefix = false;
	mbRepnePrefix = false;
	pszSegmentOverride = NULL;
	rex = 0;

	do {
		s = ApplyRuleset(pRuleBase + pRuleSystemOffsets[0], stack, mpHeap, src2, bytes, src_end);

		bytes -= (src_end - src2);
		src2 = src_end;
	} while(!*s && bytes);

	if (!*s)
		return "???";

	count = src2 - source;

	char *dst = s, *dst_base;

	while(*dst++)
		;

	dst_base = dst;

	char *t = s;
	while(*t && *t != ' ')
		++t;

	if (*t)
		*t++ = 0;

	sprintf(dst, "%-6s %s", s, t);

	return dst_base;
}

void *VDDisassemblyContext::Decompress(void *_dst, const unsigned char *src, int /*src_len*/) {
	unsigned char *dst = (unsigned char *)_dst;

	// read ruleset count
	int rulesets = *src++;
	unsigned *prstab = (unsigned *)dst;

	dst += sizeof(unsigned) * (rulesets + 1);

	// decompress rulesets sequentially
	for(int rs=0; rs<rulesets; ++rs) {
		prstab[rs+1] = dst - (unsigned char *)_dst;

		const unsigned char *pattern_cache[4][2];
		const unsigned char *result_cache[4][2];

		while(src[0] || src[1]) {
			unsigned char packctl;
			int packsrc, cnt;

			// read pack control byte and copy prematch-literal-postmatch for pattern
			packctl = *src++;
			packsrc = packctl >> 6;

			int prematch = (packctl & 7) * 2;
			int postmatch = ((packctl>>3) & 7) * 2;
			int literal = (*src++ - 1) * 2;

			*dst++ = (unsigned char)(literal + prematch + postmatch);

			const unsigned char *pattern_start = dst;

			for(cnt=0; cnt<prematch; ++cnt)
				*dst++ = pattern_cache[packsrc][0][cnt];

			for(cnt=0; cnt<literal; ++cnt)
				*dst++ = *src++;

			for(cnt=0; cnt<postmatch; ++cnt)
				*dst++ = pattern_cache[packsrc][1][cnt-postmatch];

			// cycle pattern cache

			for(cnt=3; cnt>0; --cnt) {
				pattern_cache[cnt][0] = pattern_cache[cnt-1][0];
				pattern_cache[cnt][1] = pattern_cache[cnt-1][1];
			}

			pattern_cache[0][0] = pattern_start;
			pattern_cache[0][1] = dst;

			// read pack control byte and copy prematch-literal-postmatch for result

			packctl = *src++;
			packsrc = packctl >> 6;

			prematch = (packctl & 7);
			postmatch = ((packctl>>3) & 7);
			literal = (*src++ - 1);

			*dst++ = (unsigned char)(prematch + postmatch + literal);

			const unsigned char *result_start = dst;

			for(cnt=0; cnt<prematch; ++cnt)
				*dst++ = result_cache[packsrc][0][cnt];

			for(cnt=0; cnt<literal; ++cnt)
				*dst++ = *src++;

			for(cnt=0; cnt<postmatch; ++cnt)
				*dst++ = result_cache[packsrc][1][cnt-postmatch];

			// cycle result cache

			for(cnt=3; cnt>0; --cnt) {
				result_cache[cnt][0] = result_cache[cnt-1][0];
				result_cache[cnt][1] = result_cache[cnt-1][1];
			}

			result_cache[0][0] = result_start;
			result_cache[0][1] = dst;
		}

		src += 2;

		*dst++ = 0;
		*dst++ = 0;
	}

	prstab[0] = prstab[rulesets];

	return dst;
}

char *VDDisassemblyContext::ApplyRuleset(const unsigned char *rs, ptrdiff_t *sp, char *hp, const unsigned char *source, int bytes, const unsigned char *&source_end) {
	char *hpr;

	while(rs[0] || rs[1]) {
		const unsigned char *src_end;
		const unsigned char *result = rs + rs[0] + 1;
		const unsigned char *match_next = result + result[0] + 1;

		hpr = MatchRule(source, rs+1, rs[0]>>1, bytes, sp, hp, src_end);

		if (hpr) {
			ExpandRule(hpr, result, sp, src_end);

			source_end = src_end;
			return hpr;
		}

		rs = match_next;
	}

	return NULL;
}

char *VDDisassemblyContext::MatchRule(const unsigned char *source, const unsigned char *pattern, int pattern_len, int bytes, ptrdiff_t *sp, char *hp, const unsigned char *&source_end) {
	while(pattern_len) {
		if (!pattern[1] && pattern[0]) {
			if (pattern[0] & 0x80) {
				int count = pattern[0] & 0x3f;

				if (pattern[0] & 0x40) {
					--source;
					++bytes;
				}
			
				const unsigned char *src_end;

				hp = ApplyRuleset(pRuleBase + pRuleSystemOffsets[count+1], sp, hp, source, bytes, src_end);

				if (!hp)
					return NULL;

				*sp++ = *source;
				*sp++ = (ptrdiff_t)hp;

				while(*hp++)
					;

				bytes -= (src_end - source);
				source = src_end;
			} else if (pattern[0] < 16) {
				if (pattern[0] > bytes)
					return NULL;

				for(int i=0; i<pattern[0]; ++i) {
					*sp++ = *source++;
				}

				bytes -= pattern[0];
			} else {
				switch(pattern[0]) {
				case 16:	if (!mbSizeOverride)		return NULL;	break;
				case 17:	if (!mbAddressOverride)	return NULL;	break;
				case 18:	if (!mbRepnePrefix)		return NULL;	break;
				case 19:	if (!mbRepePrefix)			return NULL;	break;
				case 20:	if (pszSegmentOverride)	return NULL;	break;
				case 21:	if (!(rex & 8))			return NULL;	break;
				}
			}
		} else {
			if (!bytes)
				return NULL;

			unsigned char b = *source++;

			if ((b & pattern[1]) != pattern[0])
				return NULL;

			*sp++ = b;
			--bytes;
		}
		pattern += 2;
		--pattern_len;
	}

	if (!pattern_len) {
		source_end = source;
		return hp;
	}

	return NULL;
}

void VDDisassemblyContext::ExpandRule(char *s, const unsigned char *result, const ptrdiff_t *sp_base, const unsigned char *source) {
	static const char *const reg64[16]={"rax","rcx","rdx","rbx","rsp","rbp","rsi","rdi","r8" ,"r9" ,"r10" ,"r11" ,"r12" ,"r13" ,"r14" ,"r15" };
	static const char *const reg32[16]={"eax","ecx","edx","ebx","esp","ebp","esi","edi","r8d","r9d","r10d","r11d","r12d","r13d","r14d","r15d"};
	static const char *const reg16[16]={ "ax", "cx", "dx", "bx", "sp", "bp", "si", "di","r8w","r9w","r10w","r11w","r12w","r13w","r14w","r15w"};
	static const char *const reg8a[16]={ "al"," cl", "dl", "bl","spl","bpl","sil","dil","r8b","r9b","r10b","r11b","r12b","r13b","r14b","r15b"};
	static const char *const reg8 [16]={"al","cl","dl","bl","ah","ch","dh","bh"};
	static const char *const regmmx[8]={"mm0","mm1","mm2","mm3","mm4","mm5","mm6","mm7"};
	static const char *const regxmm[16]={"xmm0","xmm1","xmm2","xmm3","xmm4","xmm5","xmm6","xmm7","xmm8","xmm9","xmm10","xmm11","xmm12","xmm13","xmm14","xmm15"};
	static const char *const regcrn[8]={"cr0","cr1","cr2","cr3","cr4","cr5","cr6","cr7"};
	static const char *const regdrn[8]={"dr0","dr1","dr2","dr3","dr4","dr5","dr6","dr7"};
	static const char *const regseg[8]={"es","cs","ss","ds","fs","gs","?6s","?7s"};
	static const char *const regf[8]={"st(0)","st(1)","st(2)","st(3)","st(4)","st(5)","st(6)","st(7)"};

	static const char *const *const sStaticLabels[]={
		reg32,
		reg16,
		reg8,
		regmmx,
		regxmm,
		regcrn,
		regdrn,
		regseg,
		regf
	};

	const unsigned char *result_limit = result + result[0]+1;

	++result;

	while(result < result_limit) {
		char c = *result++;

		if ((unsigned char)(c&0x7f) < 32) {
			if (c & 0x80) {
				c &= 0x7f;
				*s++ = ' ';
			}

			static const unsigned char static_bitfields[8]={
				0x80, 0x30, 0x33, 0x00, 0x00, 0x00, 0x00, 0x00
			};

			unsigned char control_byte = (unsigned char)*result++;
			unsigned char bitfield = static_bitfields[control_byte >> 5];

			if (!bitfield)
				bitfield = (unsigned char)*result++;

			int bf_start = bitfield & 15;
			int bf_siz = bitfield>>4;
			const char *arg_s = (const char *)sp_base[c-1];
			int arg = (sp_base[c-1] >> bf_start) & ((1<<bf_siz)-1);

			control_byte &= 0x1f;

			if (control_byte < 10) {
				s = strtack(s, sStaticLabels[control_byte-1][arg]);
			} else {
				long symoffset = 0;

				switch(control_byte) {
				case kTarget_r1632:
					s = strtack(s, reg32[arg] + mbSizeOverride);
					break;
				case kTarget_rmx:
					s = strtack(s, (mbSizeOverride ? regxmm : regmmx)[arg]);
					break;
				case kTarget_lx:
					symoffset = VDDisasmPack32(sp_base + c - 1);
					s += sprintf(s, "%08lx", symoffset);
					break;
				case kTarget_hx:
					s += sprintf(s, "%02x%02x"
								, (unsigned char)sp_base[c]
								, (unsigned char)sp_base[c-1]
								);
					break;
				case kTarget_x:
					s += sprintf(s, "0%02xh" + ((unsigned char)arg < 0xa0), arg);
					break;
				case kTarget_lo:
					symoffset = VDDisasmPack32(sp_base + c - 1);
					s += sprintf(s, "%c%02lx", symoffset<0 ? '-' : '+', abs(symoffset));
					break;
				case kTarget_ho:
					{
						int x =  ((unsigned char)sp_base[c  ] << 8)
								+ (unsigned char)sp_base[c-1];

						s += sprintf(s, "%c%02lx", x<0 ? '-' : '+', abs(x));
					}
					break;
				case kTarget_o:
					s += sprintf(s, "%c%02xh", arg&0x80?'-':'+', abs((signed char)arg));
					break;
				case kTarget_la:
					symoffset = (long)source + VDDisasmPack32(sp_base + c - 1) + physToVirtOffset;
					s += sprintf(s, "%08lx", symoffset);
					break;
				case kTarget_ha:
					symoffset = (long)source + (signed short)(sp_base[c-1] + (sp_base[c]<<8)) + physToVirtOffset;
					s += sprintf(s, "%08lx", symoffset);
					break;
				case kTarget_a:
					symoffset = (long)source + (signed char)arg + physToVirtOffset;
					s += sprintf(s, "%08lx", symoffset);
					break;
				case kTarget_s:
					s = strtack(s, arg_s);
					break;
				case kTarget_r3264:
					s = strtack(s, (rex & 8 ? reg32 : reg64)[arg + ((rex & 0x04) << 1)]);
					break;
				case kTarget_r163264:
					s = strtack(s, (rex & 8 ? reg64 : mbAddressOverride ? reg16 : reg32)[arg + ((rex & 0x04) << 1)]);
					break;
				case kTarget_ext:
					switch(*result++) {
					case kTarget_ext_r3264rexX:
						s = strtack(s, (mbAddressOverride ? reg32 : reg64)[arg + ((rex & 0x02) << 2)]);
						break;
					case kTarget_ext_r3264rexB:
						s = strtack(s, (mbAddressOverride ? reg32 : reg64)[arg + ((rex & 0x01) << 3)]);
						break;
					}
					break;
				}

				if (symoffset && mpSymLookup) {
					symoffset = mpSymLookup(this, (unsigned long)symoffset, s+2, 128);

					if (symoffset >= 0) {
						s[0] = ' ';
						s[1] = '(';
						s += 2;
						while(*s)
							++s;
						if (symoffset)
							s += sprintf(s, "+%02x", symoffset);
						*s++ = ')';
					}
				}
			}
		} else if ((unsigned char)c >= 0xe0) {
			switch(c) {
			case kTarget_ap:
				if (pszSegmentOverride) {
					s = strtack(s, pszSegmentOverride);
					*s++ = ':';
				}
				break;
			case kTarget_p_cs:	pszSegmentOverride = regseg[1];	break;
			case kTarget_p_ss:	pszSegmentOverride = regseg[2];	break;
			case kTarget_p_ds:	pszSegmentOverride = regseg[3];	break;
			case kTarget_p_es:	pszSegmentOverride = regseg[0];	break;
			case kTarget_p_fs:	pszSegmentOverride = regseg[4];	break;
			case kTarget_p_gs:	pszSegmentOverride = regseg[5];	break;
			case kTarget_p_66:	mbSizeOverride = true;				break;
			case kTarget_p_67:	mbAddressOverride = true;			break;
			case kTarget_p_F2:	mbRepnePrefix = true;				break;
			case kTarget_p_F3:	mbRepePrefix = true;				break;
			case kTarget_p_rex:	rex = (unsigned char)sp_base[0];	break;
			}
		} else
			*s++ = c;
	}

	*s = 0;
}
